import errno
import logging
import os
import re
from shutil import copyfile
import sys

import yaml

from django.apps import apps
from django.conf import settings
from django.db.utils import OperationalError, ProgrammingError
from django.utils.encoding import force_str
from django.utils.functional import Promise
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.serialization import yaml_dump, yaml_load

from .exceptions import SettingsException, SettingsExceptionRevert
from .literals import (
    COMMAND_NAME_SETTINGS_REVERT, NAMESPACE_VERSION_INITIAL,
    SMART_SETTINGS_NAMESPACES_NAME
)

logger = logging.getLogger(name=__name__)


def read_configuration_file(filepath):
    try:
        with open(file=filepath) as file_object:
            file_object.seek(0, os.SEEK_END)
            if file_object.tell():
                file_object.seek(0)
                try:
                    return yaml_load(stream=file_object)
                except yaml.YAMLError as exception:
                    exit(
                        'Error loading configuration file: {}; {}'.format(
                            filepath, exception
                        )
                    )
    except IOError as exception:
        if exception.errno == errno.ENOENT:
            # No config file, return empty dictionary.
            return {}
        else:
            raise


class SettingCluster(AppsModuleLoaderMixin):
    _loader_module_name = 'settings'

    def __init__(self, name):
        self.configuration_file_cache = None
        self.name = name
        self.namespace_dict = {}
        self.setting_dict = {}

    def do_cache_invalidate(self):
        for namespace in self.namespace_dict.values():
            namespace.do_cache_invalidate()

        self.configuration_file_cache = None

    def do_configuration_file_revert(self):
        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            try:
                copyfile(
                    src=settings.CONFIGURATION_LAST_GOOD_FILEPATH,
                    dst=settings.CONFIGURATION_FILEPATH
                )
            except IOError as exception:
                if exception.errno == errno.ENOENT:
                    raise SettingsException(
                        'There is no last valid version to restore.'
                    ) from exception
                else:
                    raise
        else:
            logger.info(
                'Local storage is disabled, cannot revert not existing '
                'configuration.'
            )

    def do_configuration_file_save(self, path=None):
        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            if not path:
                path = settings.CONFIGURATION_FILEPATH

            try:
                with open(file=path, mode='w') as file_object:
                    file_object.write(
                        self.get_data_dump()
                    )
            except IOError as exception:
                if exception.errno == errno.ENOENT:
                    logger.warning(
                        'The path to the configuration file `%s` doesn\'t '
                        'exist. It is not possible to save the backup file.',
                        path
                    )
        else:
            logger.info(
                'Local storage is disabled, skip saving configuration.'
            )

    def do_last_known_good_save(self):
        # Don't write over the last good configuration if we are trying
        # to restore the last good configuration.
        if COMMAND_NAME_SETTINGS_REVERT not in sys.argv and not settings.CONFIGURATION_FILE_IGNORE:
            self.do_configuration_file_save(
                path=settings.CONFIGURATION_LAST_GOOD_FILEPATH
            )

    def do_namespace_add(self, **kwargs):
        setting_namespace = SettingNamespace(cluster=self, **kwargs)

        if setting_namespace.name in self.namespace_dict:
            raise SettingsException(
                'Setting namespace "%s" already exists in '
                'cluster.' % setting_namespace.name
            )

        self.namespace_dict[setting_namespace.name] = setting_namespace
        return setting_namespace

    def do_namespace_remove(self, name):
        setting_namespace = self.get_namespace(name=name)

        self.namespace_dict.pop(setting_namespace.name)

    def do_post_edit_function_call(self):
        ContentType = apps.get_model(
            app_label='contenttypes', model_name='ContentType'
        )

        for namespace in self.namespace_dict.values():
            namespace.do_post_edit_function_call()

        # Clear the content type cache to avoid the event system from trying
        # to use the same ID that were cached when the setting post edit
        # functions executed. This is because the settings execute before
        # the apps objects are created.
        ContentType.objects.clear_cache()

    def do_settings_updated_clear(self):
        UpdatedSetting = apps.get_model(
            app_label='smart_settings', model_name='UpdatedSetting'
        )

        queryset = UpdatedSetting.objects.all()

        try:
            queryset.delete()
        except (OperationalError, ProgrammingError):
            """
            Non fatal. Non initialized installation. Ignore exception.
            """

    def get_configuration_file_content(self):
        if settings.CONFIGURATION_FILE_IGNORE:
            return {}
        else:
            # Cache content the of the configuration file to speed up
            # initial boot up.
            if not self.configuration_file_cache:
                self.configuration_file_cache = read_configuration_file(
                    filepath=settings.CONFIGURATION_FILEPATH
                ) or {}
            return self.configuration_file_cache

    def get_data_dump(self, filter_term=None, namespace_name=None):
        UpdatedSetting = apps.get_model(
            app_label='smart_settings', model_name='UpdatedSetting'
        )

        dictionary = {}

        if not namespace_name:
            namespace_dictionary = {}
            for namespace in self.get_namespace_list():
                namespace_dictionary[namespace.name] = {
                    'version': namespace.version
                }

            dictionary[SMART_SETTINGS_NAMESPACES_NAME] = namespace_dictionary

        if namespace_name:
            namespace_list = (
                self.get_namespace(name=namespace_name),
            )
        else:
            namespace_list = self.get_namespace_list()

        for namespace in namespace_list:
            for setting in namespace.get_setting_list():
                # If a namespace is specified, filter the list by that
                # namespace otherwise return always True to include all
                # (or not None == True).
                if (filter_term and filter_term.lower() in setting.global_name.lower()) or not filter_term:
                    try:
                        updated_setting = UpdatedSetting.objects.get(
                            global_name=setting.global_name
                        )
                    except (OperationalError, ProgrammingError, UpdatedSetting.DoesNotExist):
                        expressed_value = Setting.express_promises(
                            value=setting.value
                        )
                        dictionary[setting.global_name] = expressed_value
                    else:
                        dictionary[setting.global_name] = updated_setting.value_new

        return yaml_dump(data=dictionary, default_flow_style=False)

    def get_is_changed(self):
        UpdatedSetting = apps.get_model(
            app_label='smart_settings', model_name='UpdatedSetting'
        )

        try:
            return UpdatedSetting.objects.exists()
        except (OperationalError, ProgrammingError):
            return False

    def get_namespace(self, name):
        return self.namespace_dict[name]

    def get_namespace_configuration(self, name):
        namespace_configuration_map = self.get_namespace_configuration_map()

        return namespace_configuration_map.get(
            name, {}
        )

    def get_namespace_configuration_map(self):
        configuration_file_content = self.get_configuration_file_content()

        return configuration_file_content.get(
            SMART_SETTINGS_NAMESPACES_NAME, {}
        )

    def get_namespace_list(self):
        return sorted(
            self.namespace_dict.values(), key=lambda x: x.label
        )

    def get_setting(self, global_name):
        return self.setting_dict[global_name]

    def get_setting_list(self):
        return self.setting_dict.values()


class SettingNamespace:
    def __init__(
        self, cluster, name, label, migration_class=None,
        version=NAMESPACE_VERSION_INITIAL
    ):
        self.cluster = cluster
        self.migration_class = migration_class
        self.name = name
        self.label = label
        self.setting_dict = {}
        self.version = version

    def __str__(self):
        return force_str(s=self.label)

    def do_cache_invalidate(self):
        for setting in self.setting_dict.values():
            setting.do_cache_invalidate()

    def do_post_edit_function_call(self):
        for setting in self.setting_dict.values():
            setting.do_post_edit_function_call()

    def do_migrate(self, setting):
        if self.migration_class:
            self.migration_class(namespace=self).do_migrate(setting=setting)

    def do_setting_add(self, **kwargs):
        setting = Setting(namespace=self, **kwargs)

        if setting.global_name in self.setting_dict:
            raise SettingsException(
                'Setting "%s" already exists in '
                'namespace.' % setting.global_name
            )

        self.setting_dict[setting.global_name] = setting
        self.cluster.setting_dict[setting.global_name] = setting

        return setting

    def do_setting_remove(self, global_name):
        setting = self.setting_dict.get(global_name)

        self.setting_dict.pop(setting.global_name)
        self.cluster.setting_dict.pop(setting.global_name)

        return setting

    def get_configuration_file_version(self):
        return self.cluster.get_namespace_configuration(name=self.name).get(
            'version', NAMESPACE_VERSION_INITIAL
        )

    def get_setting(self, global_name):
        return self.setting_dict[global_name]

    def get_setting_list(self):
        return sorted(
            self.setting_dict.values(), key=lambda x: x.global_name
        )


SettingNamespace.verbose_name = _(message='Settings namespace')


class SettingNamespaceMigration:
    @staticmethod
    def get_method_name(setting):
        return setting.global_name.lower()

    def __init__(self, namespace):
        self.namespace = namespace

    def do_migrate(self, setting):
        if self.namespace.get_configuration_file_version() != self.namespace.version:
            setting_method_name = SettingNamespaceMigration.get_method_name(
                setting=setting
            )

            # Get methods for this setting.
            pattern = r'{}_\d{{4}}'.format(setting_method_name)
            setting_methods = re.findall(
                pattern=pattern, string='\n'.join(
                    dir(self)
                )
            )

            # Get order of execution of setting methods.
            version_list = [
                method.replace(
                    '{}_'.format(setting_method_name), ''
                ) for method in setting_methods
            ]
            try:
                start = version_list.index(
                    self.namespace.get_configuration_file_version()
                )
            except ValueError:
                start = 0

            try:
                end = version_list.index(self.namespace.version)
            except ValueError:
                end = None

            value = setting.value_raw
            for version in version_list[start:end]:
                method = getattr(
                    self, self.get_method_name_full(
                        setting=setting, version=version
                    ), None
                )
                if method:
                    value = method(value=value)

            setting.value_raw = value

    def get_method_name_full(self, setting, version):
        return '{}_{}'.format(
            self.get_method_name(setting=setting), version
        )


class Setting:
    @staticmethod
    def deserialize_value(value):
        return yaml_load(stream=value)

    @staticmethod
    def express_promises(value):
        """
        Walk all the elements of a value and force promises to text.
        """
        if isinstance(value, (list, tuple)):
            return [
                Setting.express_promises(item) for item in value
            ]
        elif isinstance(value, Promise):
            return force_str(s=value)
        else:
            return value

    @staticmethod
    def serialize_value(value):
        result = yaml_dump(
            allow_unicode=True, data=Setting.express_promises(value=value),
            default_flow_style=False
        )
        # safe_dump returns bytestrings.
        # Disregard the last 3 dots that mark the end of the YAML document.
        if force_str(s=result).endswith('...\n'):
            result = result[:-4]

        return result

    def __init__(
        self, namespace, global_name, default, choices=None, help_text=None,
        is_path=False, post_edit_function=None, validation_function=None
    ):
        self.choices = choices
        self.default = default
        self.help_text = help_text
        self.environment_variable = False
        self.global_name = global_name
        self.loaded = False
        self.namespace = namespace
        self.post_edit_function = post_edit_function
        self.validation_function = validation_function
        self.value_raw_new = None

    def __str__(self):
        return str(self.global_name)

    def do_cache_invalidate(self):
        self.loaded = False

    def do_migrate(self):
        self.namespace.do_migrate(setting=self)

    def do_post_edit_function_call(self):
        if self.post_edit_function:
            try:
                self.post_edit_function(setting=self)
            except Exception as exception:
                raise SettingsException(
                    'Unable to execute setting post update function '
                    'for setting "{}". Verify the value of the setting or '
                    'rollback to the previous known working configuration '
                    'file.'.format(self.global_name)
                ) from exception

    def do_value_cache(self, global_name=None, default_override=None):
        global_name = global_name or self.global_name

        environment_value = os.environ.get(
            'MAYAN_{}'.format(global_name)
        )
        if environment_value:
            self.environment_variable = True
            try:
                self.value_raw = yaml_load(stream=environment_value)
            except yaml.YAMLError as exception:
                raise type(exception)(
                    'Error interpreting environment variable: {} with '
                    'value: {}; {}'.format(
                        global_name, environment_value, exception
                    )
                )
        else:
            try:
                # Try the config file.
                configuration_file_content = self.namespace.cluster.get_configuration_file_content()
                self.value_raw = configuration_file_content[global_name]
            except KeyError:
                try:
                    # Try the Django settings variable.
                    self.value_raw = getattr(
                        settings, global_name
                    )
                except AttributeError:
                    # Finally set to the default value.
                    if default_override:
                        self.value_raw = default_override
                    else:
                        self.value_raw = self.default
            else:
                # Found in the config file, try to migrate the value.
                self.do_migrate()

        if self.validation_function:
            self.value_raw = self.validation_function(
                raw_value=self.value_raw, setting=self
            )

        self.value_yaml = Setting.serialize_value(value=self.value_raw)
        self.loaded = True

    def get_value_choices(self):
        return self.choices

    get_value_choices.short_description = _(message='Choices')
    get_value_choices.help_text = _(
        'Possible values allowed for this setting.'
    )

    def do_value_raw_set(self, raw_value):
        self.value = Setting.serialize_value(value=raw_value)
        self.loaded = True

    def do_value_raw_validate(self, raw_value):
        if self.validation_function:
            return self.validation_function(
                raw_value=raw_value, setting=self
            )

    def do_value_revert(self):
        UpdatedSetting = apps.get_model(
            app_label='smart_settings', model_name='UpdatedSetting'
        )

        if not self.get_has_value_new():
            raise SettingsExceptionRevert(
                _(
                    'Cannot revert setting. Setting value has not been '
                    'updated.'
                )
            )

        updated_setting = UpdatedSetting.objects.get(
            global_name=self.global_name
        )

        self.do_value_set(value=updated_setting.value_old)
        updated_setting.delete()

    def do_value_set(self, value):
        UpdatedSetting = apps.get_model(
            app_label='smart_settings', model_name='UpdatedSetting'
        )

        raw_value = Setting.deserialize_value(value=value)

        self.value_raw_new = raw_value

        if self.value_raw_new != self.value:
            updated_setting, created = UpdatedSetting.objects.update_or_create(
                global_name=self.global_name, defaults={
                    'value_new': self.value_raw_new,
                    'value_old': self.value
                }
            )
        else:
            queryset = UpdatedSetting.objects.filter(
                global_name=self.global_name
            )
            queryset.delete()

    def get_default(self):
        return Setting.serialize_value(value=self.default)

    get_default.short_description = _(message='Default')

    def get_has_value_new(self):
        UpdatedSetting = apps.get_model(
            app_label='smart_settings', model_name='UpdatedSetting'
        )

        queryset = UpdatedSetting.objects.filter(
            global_name=self.global_name
        )

        return queryset.exists()

    get_has_value_new.short_description = _(message='Modified')
    get_has_value_new.help_text = _(
        'The value of this setting being modified since the last restart.'
    )

    def get_is_overridden(self):
        return self.environment_variable

    get_is_overridden.short_description = _(message='Overridden')
    get_is_overridden.help_text = _(
        'The value of the setting is being overridden by an environment '
        'variable.'
    )

    def get_value_current(self):
        UpdatedSetting = apps.get_model(
            app_label='smart_settings', model_name='UpdatedSetting'
        )

        has_value_new = self.get_has_value_new()

        if has_value_new:
            updated_setting = UpdatedSetting.objects.get(
                global_name=self.global_name
            )

            return Setting.serialize_value(value=updated_setting.value_new)
        else:
            return self.serialized_value

    @property
    def pk(self):
        """
        Compatibility property for views that expect model instances.
        """
        return self.global_name

    @property
    def serialized_value(self):
        """
        YAML serialize value of the setting.
        Used for UI display.
        """
        if not self.loaded:
            self.do_value_cache()

        return self.value_yaml

    @property
    def value(self):
        if not self.loaded:
            self.do_value_cache()

        return self.value_raw

    @value.setter
    def value(self, value):
        # value is in YAML format.
        self.value_yaml = value
        self.value_raw = Setting.deserialize_value(value=self.value_yaml)



from django.utils.translation import gettext_lazy as _

COMMAND_NAME_SEARCH_INDEX_OBJECTS = 'search_index_objects'
COMMAND_NAME_SEARCH_REINDEX = 'search_reindex'
COMMAND_NAME_SEARCH_STATUS = 'search_status'

DEFAULT_SEARCH_BACKEND = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'
DEFAULT_SEARCH_BACKEND_ARGUMENTS = {}
DEFAULT_SEARCH_DISABLE_SIMPLE_SEARCH = False
DEFAULT_SEARCH_INDEXING_CHUNK_SIZE = 25
DEFAULT_SEARCH_MATCH_ALL_DEFAULT_VALUE = 'False'
DEFAULT_SEARCH_QUERY_RESULTS_LIMIT = 10000
DEFAULT_SEARCH_RESULTS_LIMIT = 1000
DEFAULT_TEST_SEARCH_BACKEND = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'

ERROR_TEXT_NO_RESULT_SCOPE = 'No result scope has been specified.'

FILTER_PREFIX = 'filter_'

MATCH_ALL_FIELD_CHOICES = (
    (True, _(message='Yes')),
    (False, _(message='No'))
)
MATCH_ALL_FIELD_NAME = '_match_all'
MATCH_ALL_VALUES = ('on', 'true', 'yes')
MESSAGE_FEATURE_NO_STATUS = 'This backend does not provide status information.'

QUERY_PARAMETER_ANY_FIELD = 'q'

SCOPE_DELIMITER = '_'
SCOPE_MARKER = '__'
SCOPE_RESULT_MARKER = 'result'

SCOPE_OPERATOR_AND = 'AND'
SCOPE_OPERATOR_NOT = 'NOT'
SCOPE_OPERATOR_OR = 'OR'


def scope_operation_and(*args):
    result = set(
        args[0]
    )
    for argument in args[1:]:
        result = result.intersection(argument)

    return result


def scope_operation_not(*args):
    result = set(
        args[0]
    )
    for argument in args[1:]:
        result = result.difference(argument)

    return result


def scope_operation_or(*args):
    result = set(
        args[0]
    )
    for argument in args[1:]:
        result = result.union(argument)

    return result


SCOPE_OPERATOR_CHOICES = {
    SCOPE_OPERATOR_AND: scope_operation_and,
    SCOPE_OPERATOR_NOT: scope_operation_not,
    SCOPE_OPERATOR_OR: scope_operation_or
}

DEFAULT_SEARCH_DEFAULT_OPERATOR = SCOPE_OPERATOR_AND

SEARCH_MODEL_NAME_KWARG = 'search_model_pk'

TASK_DEINDEX_INSTANCE_MAX_RETRIES = 40
TASK_DEINDEX_INSTANCE_RETRY_BACKOFF_MAX = 60

TASK_INDEX_INSTANCE_MAX_RETRIES = 40
TASK_INDEX_INSTANCE_RETRY_BACKOFF_MAX = 60

TASK_INDEX_INSTANCES_MAX_RETRIES = 40
TASK_INDEX_INSTANCES_RETRY_BACKOFF_MAX = 60

TASK_INDEX_RELATED_INSTANCE_M2M_MAX_RETRIES = 40
TASK_INDEX_RELATED_INSTANCE_M2M_RETRY_BACKOFF_MAX = 60

TERM_OPERATOR_AND = 'AND'
TERM_OPERATOR_NOT = 'NOT'
TERM_OPERATOR_OR = 'OR'
TERM_OPERATORS = [
    TERM_OPERATOR_AND, TERM_OPERATOR_NOT, TERM_OPERATOR_OR
]

TERM_MARKER_QUOTE = '"'
TERM_MARKER_RAW = '`'
TERM_MARKER_SPACE_CHARACTER = ' '



import logging

from django.apps import apps
from django.core.exceptions import ObjectDoesNotExist

from mayan.apps.lock_manager.exceptions import LockError
from mayan.celery import app

from .exceptions import DynamicSearchException, DynamicSearchRetry
from .literals import (
    TASK_DEINDEX_INSTANCE_MAX_RETRIES,
    TASK_DEINDEX_INSTANCE_RETRY_BACKOFF_MAX,
    TASK_INDEX_INSTANCE_MAX_RETRIES, TASK_INDEX_INSTANCE_RETRY_BACKOFF_MAX,
    TASK_INDEX_INSTANCES_MAX_RETRIES,
    TASK_INDEX_INSTANCES_RETRY_BACKOFF_MAX,
    TASK_INDEX_RELATED_INSTANCE_M2M_MAX_RETRIES,
    TASK_INDEX_RELATED_INSTANCE_M2M_RETRY_BACKOFF_MAX
)
from .search_backends import SearchBackend
from .search_models import SearchModel

logger = logging.getLogger(name=__name__)


@app.task(
    bind=True, ignore_result=True,
    max_retries=TASK_DEINDEX_INSTANCE_MAX_RETRIES, retry_backoff=True,
    retry_backoff_max=TASK_DEINDEX_INSTANCE_RETRY_BACKOFF_MAX
)
def task_deindex_instance(self, app_label, model_name, object_id):
    logger.info('Executing')

    Model = apps.get_model(app_label=app_label, model_name=model_name)
    instance = Model._meta.default_manager.get(pk=object_id)

    try:
        SearchBackend.get_instance().deindex_instance(instance=instance)
    except (DynamicSearchRetry, LockError) as exception:
        raise self.retry(exc=exception)
    except ObjectDoesNotExist as exception:
        # Object was deleted before it could be deindexed.
        logger.info(
            str(
                exception
            )
        )

    logger.info('Finished')


@app.task(
    bind=True, ignore_result=True,
    max_retries=TASK_INDEX_INSTANCE_MAX_RETRIES, retry_backoff=True,
    retry_backoff_max=TASK_INDEX_INSTANCE_RETRY_BACKOFF_MAX
)
def task_index_instance(
    self, app_label, model_name, object_id, exclude_app_label=None,
    exclude_model_name=None, exclude_kwargs=None
):
    logger.info('Executing')

    Model = apps.get_model(app_label=app_label, model_name=model_name)
    if exclude_app_label and exclude_model_name:
        ExcludeModel = apps.get_model(
            app_label=exclude_app_label, model_name=exclude_model_name
        )
    else:
        ExcludeModel = None

    try:
        instance = Model._meta.default_manager.get(pk=object_id)
    except Model.DoesNotExist as exception:
        raise self.retry(exc=exception)

    try:
        SearchBackend.get_instance().index_instance(
            exclude_kwargs=exclude_kwargs, exclude_model=ExcludeModel,
            instance=instance
        )
    except (DynamicSearchRetry, LockError) as exception:
        raise self.retry(exc=exception)
    except ObjectDoesNotExist as exception:
        # Object was deleted before it could be indexed.
        logger.info(
            str(
                exception
            )
        )
    except Exception as exception:
        kwargs = {
            'app_label': app_label,
            'model_name': model_name,
            'object_id': object_id,
            'exclude_app_label': exclude_app_label,
            'exclude_model_name': exclude_model_name,
            'exclude_kwargs': exclude_kwargs
        }
        error_message = (
            'Unexpected error calling `task_index_instance` with keyword '
            'arguments {}.'
        ).format(kwargs)

        logger.error(error_message)
        raise DynamicSearchException(error_message) from exception
    else:
        logger.info('Finished')


@app.task(
    bind=True, ignore_result=True,
    max_retries=TASK_INDEX_INSTANCES_MAX_RETRIES, retry_backoff=True,
    retry_backoff_max=TASK_INDEX_INSTANCES_RETRY_BACKOFF_MAX
)
def task_index_instances(self, search_model_full_name, id_list):
    search_model = SearchModel.get(name=search_model_full_name)

    kwargs = {
        'id_list': id_list, 'search_model': search_model
    }

    try:
        SearchBackend.get_instance().index_instances(**kwargs)
    except (DynamicSearchRetry, LockError) as exception:
        raise self.retry(exc=exception)
    except Exception as exception:
        error_message = (
            'Unexpected error calling `task_index_instances` with '
            'keyword arguments {}.'
        ).format(kwargs)

        logger.error(error_message)
        raise DynamicSearchException(error_message) from exception


@app.task(
    bind=True, ignore_result=True,
    max_retries=TASK_INDEX_RELATED_INSTANCE_M2M_MAX_RETRIES,
    retry_backoff=True,
    retry_backoff_max=TASK_INDEX_RELATED_INSTANCE_M2M_RETRY_BACKOFF_MAX
)
def task_index_related_instance_m2m(
    self, action, instance_app_label, instance_model_name,
    instance_object_id, model_app_label, model_model_name, pk_set,
    serialized_search_model_related_paths
):
    InstanceModel = apps.get_model(
        app_label=instance_app_label, model_name=instance_model_name
    )
    instance = InstanceModel.objects.get(pk=instance_object_id)

    Model = apps.get_model(
        app_label=model_app_label, model_name=model_model_name
    )

    search_model_related_paths = {}

    for key, value in serialized_search_model_related_paths.items():
        app_label, model_name = key.split('.')
        DeserializedModel = apps.get_model(
            app_label=app_label, model_name=model_name
        )
        search_model_related_paths[DeserializedModel] = value

    SearchBackend.index_related_instance_m2m(
        action=action, instance=instance, model=Model, pk_set=pk_set,
        search_model_related_paths=search_model_related_paths
    )


@app.task(ignore_result=True)
def task_reindex_backend():
    search_backend = SearchBackend.get_instance()
    search_backend.reset()

    for search_model in SearchModel.all():
        for id_list in search_model.get_id_groups():
            task_index_instances.apply_async(
                kwargs={
                    'id_list': id_list,
                    'search_model_full_name': search_model.full_name
                }
            )



from django.db import models

from mayan.apps.testing.tests.base import BaseTestCase

from ..utils import (
    ResolverPipelineModelAttribute, flatten_list, group_iterator, parse_range
)


class FlattenListTestCase(BaseTestCase):
    def test_string_values(self):
        self.assertEqual(
            list(
                flatten_list(value='test string')
            ), ['test string']
        )

        self.assertEqual(
            list(
                flatten_list(
                    value=['test string']
                )
            ), ['test string']
        )

        self.assertEqual(
            list(
                flatten_list(
                    value=['test string1', 'test string2']
                )
            ), ['test string1', 'test string2']
        )

        self.assertEqual(
            list(
                flatten_list(
                    value=['test string1', 1]
                )
            ), ['test string1', 1]
        )

        self.assertEqual(
            list(
                flatten_list(
                    value=[
                        ['test string1'], 1
                    ]
                )
            ), ['test string1', 1]
        )

        self.assertEqual(
            list(
                flatten_list(
                    value=[
                        ['test string1'], [1]
                    ]
                )
            ), ['test string1', 1]
        )


class GroupIteratorTestCase(BaseTestCase):
    def test_basic(self):
        self.assertEqual(
            list(
                group_iterator(
                    iterable=parse_range(range_string='1')
                )
            ), [
                (1)
            ]
        )

        self.assertEqual(
            list(
                group_iterator(
                    iterable=parse_range(range_string='1,5-10'), group_size=2
                )
            ), [
                (1, 5), (6, 7), (8, 9), (10,)
            ]
        )

    def test_empty_range(self):
        self.assertEqual(
            list(
                group_iterator(
                    iterable=parse_range(range_string='')
                )
            ), []
        )


class ParseRangeTestCase(BaseTestCase):
    def test_parse_range(self):
        self.assertEqual(
            list(
                parse_range(range_string='1')
            ), [1]
        )

        self.assertEqual(
            list(
                parse_range(range_string='1-5')
            ), [1, 2, 3, 4, 5]
        )

        self.assertEqual(
            list(
                parse_range(range_string='2,4,6')
            ), [2, 4, 6]
        )

        self.assertEqual(
            list(
                parse_range(range_string='2,4,6-8')
            ), [2, 4, 6, 7, 8]
        )

    def test_repeated_numbers(self):
        self.assertEqual(
            list(
                parse_range(range_string='1,2,3,1,2,3')
            ), [1, 2, 3, 1, 2, 3]
        )

    def test_reverse(self):
        self.assertEqual(
            list(
                parse_range(range_string='9-5')
            ), [9, 8, 7, 6, 5]
        )

    def test_unsorted_range(self):
        self.assertEqual(
            list(
                parse_range(range_string='9,2,4,6-8')
            ), [9, 2, 4, 6, 7, 8]
        )

    def test_empty_range(self):
        self.assertEqual(
            list(
                parse_range(range_string='')
            ), []
        )


class ResolverRelatedManagerTestCase(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.TestModelAttribute = self._create_test_model(
            fields={
                'label': models.CharField(
                    max_length=64
                )
            }, model_name='TestModelAttribute'
        )
        self.TestModelGrandParent = self._create_test_model(
            model_name='TestModelGrandParent'
        )
        self.TestModelParent = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelGrandParent'
                )
            }, model_name='TestModelParent'
        )
        self.TestModelGrandChild = self._create_test_model(
            fields={
                'parent': models.ForeignKey(
                    on_delete=models.CASCADE, related_name='children',
                    to='TestModelParent'
                ),
                'attributes': models.ManyToManyField(
                    related_name='children', to='TestModelAttribute'
                )
            }, model_name='TestModelGrandChild'
        )

        self._test_object_grandparent = self.TestModelGrandParent.objects.create()
        self._test_object_parent = self.TestModelParent.objects.create(
            parent=self._test_object_grandparent
        )
        self._test_object_grandchild = self.TestModelGrandChild.objects.create(
            parent=self._test_object_parent
        )
        self._test_object_attribute = self.TestModelAttribute.objects.create(
            label='test attribute object'
        )
        self._test_object_grandchild.attributes.add(self._test_object_attribute)

    def test_many_to_many(self):
        result = ResolverPipelineModelAttribute.resolve(
            attribute='attributes',
            obj=self._test_object_grandchild
        )

        self.assertEqual(result.count(), 1)
        self.assertEqual(result[0], self._test_object_attribute)

    def test_many_to_many_field(self):
        result = ResolverPipelineModelAttribute.resolve(
            attribute='attributes__label',
            obj=self._test_object_grandchild
        )

        self.assertEqual(
            len(result), 1
        )
        self.assertEqual(
            result[0], self._test_object_attribute.label
        )

    def test_many_to_many_field_exclude(self):
        result = ResolverPipelineModelAttribute.resolve(
            attribute='attributes__label',
            obj=self._test_object_grandchild,
            resolver_extra_kwargs={
                'exclude': {
                    'id': '{}'.format(self._test_object_attribute.pk)
                },
                'model': self.TestModelAttribute
            }
        )

        self.assertEqual(
            len(result), 0
        )

    def test_multiple_level_reverse_relation(self):
        result = ResolverPipelineModelAttribute.resolve(
            attribute='parent__parent', obj=self._test_object_grandchild,
        )

        self.assertEqual(
            len(result), 1
        )
        self.assertEqual(
            result[0].count(), 1
        )
        self.assertEqual(
            result[0][0], self._test_object_grandparent
        )

    def test_single_level_reverse_many_to_many(self):
        result = ResolverPipelineModelAttribute.resolve(
            attribute='children',
            obj=self._test_object_attribute
        )

        self.assertEqual(
            result.count(), 1
        )
        self.assertEqual(
            result[0], self._test_object_grandchild
        )

    def test_multiple_level_reverse_relation_from_many_to_many_field(self):
        result = ResolverPipelineModelAttribute.resolve(
            attribute='children__parent__parent',
            obj=self._test_object_attribute
        )

        self.assertEqual(
            len(result), 1
        )
        self.assertEqual(
            len(
                result[0]
            ), 1
        )
        self.assertEqual(
            result[0][0].count(), 1
        )
        self.assertEqual(
            result[0][0][0], self._test_object_grandparent
        )

    def test_multiple_level_relation(self):
        result = ResolverPipelineModelAttribute.resolve(
            attribute='children__children',
            obj=self._test_object_grandparent
        )

        self.assertEqual(
            len(result), 1
        )
        self.assertEqual(
            result[0].count(), 1
        )
        self.assertEqual(
            result[0][0], self._test_object_grandchild
        )

    def test_multiple_level_relation_to_many_to_many_exclude(self):
        result = ResolverPipelineModelAttribute.resolve(
            attribute='children__children__attributes',
            obj=self._test_object_grandparent,
            resolver_extra_kwargs={
                'exclude': {
                    'id': '{}'.format(self._test_object_attribute.pk)
                },
                'model': self.TestModelAttribute
            }
        )

        self.assertEqual(
            len(result), 1
        )
        self.assertEqual(
            len(
                result[0]
            ), 1
        )
        self.assertEqual(
            result[0][0].count(), 0
        )



from rest_framework.generics import get_object_or_404

from mayan.apps.acls.models import AccessControlList

from ..models.workflow_models import Workflow


class ParentObjectWorkflowTemplateAPIViewMixin:
    def get_serializer_extra_context(self):
        context = super().get_serializer_extra_context()
        context.update(
            {
                'workflow_template': self.get_workflow_template()
            }
        )
        return context

    def get_workflow_template(self, permission=None):
        queryset = self.get_workflow_template_queryset()

        if not permission:
            mayan_external_object_permission_map = getattr(
                self, 'mayan_external_object_permission_map', {}
            )

            permission = mayan_external_object_permission_map.get(
                self.request.method, None
            )

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs.get('workflow_template_id')
        )

    def get_workflow_template_queryset(self):
        return Workflow.objects.all()


class ParentObjectWorkflowTemplateStateAPIViewMixin(
    ParentObjectWorkflowTemplateAPIViewMixin
):
    def get_serializer_extra_context(self):
        context = super().get_serializer_extra_context()
        context.update(
            {
                'workflow_template_state': self.get_workflow_template_state()
            }
        )
        return context

    def get_workflow_template_state(self, permission=None):
        queryset = self.get_workflow_template_state_queryset()

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs.get(
                'workflow_template_state_id'
            )
        )

    def get_workflow_template_state_queryset(self):
        return self.get_workflow_template().states.all()


class ParentObjectWorkflowTemplateTransitionAPIViewMixin(
    ParentObjectWorkflowTemplateAPIViewMixin
):
    def get_serializer_extra_context(self):
        context = super().get_serializer_extra_context()
        context.update(
            {
                'workflow_template_transition': self.get_workflow_template_transition()
            }
        )
        return context

    def get_workflow_template_transition(self, permission=None):
        queryset = self.get_workflow_template_transition_queryset()

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs.get(
                'workflow_template_transition_id'
            )
        )

    def get_workflow_template_transition_queryset(self):
        return self.get_workflow_template().transitions.all()



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_MIRRORING_DOCUMENT_CACHE_LOOKUP_TIMEOUT,
    DEFAULT_MIRRORING_NODE_CACHE_LOOKUP_TIMEOUT
)

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Mirroring'), name='mirroring'
)

setting_document_lookup_cache_timeout = setting_namespace.do_setting_add(
    default=DEFAULT_MIRRORING_DOCUMENT_CACHE_LOOKUP_TIMEOUT,
    global_name='MIRRORING_DOCUMENT_CACHE_LOOKUP_TIMEOUT',
    help_text=_(message='Time in seconds to cache the path lookup to a document.')
)
setting_node_lookup_cache_timeout = setting_namespace.do_setting_add(
    default=DEFAULT_MIRRORING_NODE_CACHE_LOOKUP_TIMEOUT,
    global_name='MIRRORING_NODE_CACHE_LOOKUP_TIMEOUT',
    help_text=_(message='Time in seconds to cache the path lookup to an index node.')
)



import logging

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models.document_models import Document
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerMethodAfter

from .events import (
    event_document_auto_checked_in, event_document_checked_in,
    event_document_checked_out, event_document_forcefully_checked_in
)
from .exceptions import DocumentAlreadyCheckedOut
from .managers import (
    DocumentCheckoutBusinessLogicManager, DocumentCheckoutManager
)
from .model_mixins import CheckedOutDocumentBusinessLogicMixin

logger = logging.getLogger(name=__name__)


class DocumentCheckout(ExtraDataModelMixin, models.Model):
    """
    Model to store the state and information of a document checkout.
    """
    document = models.OneToOneField(
        on_delete=models.CASCADE, to=Document, related_name='checkout',
        verbose_name=_(message='Document')
    )
    checkout_datetime = models.DateTimeField(
        auto_now_add=True, verbose_name=_(message='Check out date and time')
    )
    expiration_datetime = models.DateTimeField(
        help_text=_(
            'Amount of time to hold the document checked out in minutes.'
        ), verbose_name=_(message='Check out expiration date and time')
    )
    user = models.ForeignKey(
        on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL,
        verbose_name=_(message='User')
    )
    block_new_file = models.BooleanField(
        default=True, help_text=_(
            'Do not allow new file of this document to be uploaded.'
        ), verbose_name=_(message='Block new file upload')
    )

    objects = DocumentCheckoutManager()
    business_logic = DocumentCheckoutBusinessLogicManager()

    class Meta:
        ordering = ('pk',)
        verbose_name = _(message='Document checkout')
        verbose_name_plural = _(message='Document checkouts')

    def __str__(self):
        return str(self.document)

    def clean(self):
        if self.expiration_datetime < now():
            raise ValidationError(
                message=_(
                    'Check out expiration date and time must be in '
                    'the future.'
                )
            )

    @method_event(event_manager_class=EventManagerMethodAfter)
    def delete(self, user=None):
        self._event_target = self.document
        self._event_actor = user or getattr(self, '_event_actor', None)

        if self._event_actor:
            if self._event_actor == self.user:
                self._event_type = event_document_checked_in
            else:
                self._event_type = event_document_forcefully_checked_in
        else:
            self._event_type = event_document_auto_checked_in

        return super().delete()

    def get_absolute_url(self):
        return reverse(
            viewname='checkouts:check_out_info', kwargs={
                'document_id': self.document_id
            }
        )

    def natural_key(self):
        return self.document.natural_key()
    natural_key.dependencies = ['documents.Document']

    def save(self, *args, **kwargs):
        is_new = not self.pk
        if not is_new or self.document.is_checked_out():
            raise DocumentAlreadyCheckedOut

        result = super().save(*args, **kwargs)
        if is_new:
            event_document_checked_out.commit(
                actor=self.user, target=self.document
            )

            logger.info(
                'Document "%s" checked out by user "%s"',
                self.document, self.user
            )

        return result


class CheckedOutDocument(CheckedOutDocumentBusinessLogicMixin, Document):
    class Meta:
        proxy = True



from django.utils.translation import gettext_lazy as _

from mayan.apps.converter.permissions import (
    permission_transformation_delete, permission_transformation_edit
)
from mayan.apps.navigation.classes import Link

from ..icons import (
    icon_document_version_active, icon_document_version_create,
    icon_document_version_delete, icon_document_version_edit,
    icon_document_version_list, icon_document_version_modification,
    icon_document_version_preview, icon_document_version_print,
    icon_document_version_return_document, icon_document_version_return_list,
    icon_document_version_transformation_list_clear,
    icon_document_version_transformation_list_clone
)
from ..permissions import (
    permission_document_version_create, permission_document_version_delete,
    permission_document_version_edit, permission_document_version_print,
    permission_document_version_view, permission_document_view
)

link_document_version_active = Link(
    args='resolved_object.pk',
    icon=icon_document_version_active,
    permission=permission_document_version_edit,
    text=_(message='Make active'), view='documents:document_version_active'
)
link_document_version_create = Link(
    args='resolved_object.pk', icon=icon_document_version_create,
    permission=permission_document_version_create,
    text=_(message='Create document version'),
    view='documents:document_version_create'
)
link_document_version_single_delete = Link(
    args='resolved_object.pk',
    icon=icon_document_version_delete,
    permission=permission_document_version_delete, tags='dangerous',
    text=_(message='Delete'), view='documents:document_version_single_delete'
)
link_document_version_multiple_delete = Link(
    icon=icon_document_version_delete, tags='dangerous',
    text=_(message='Delete'), view='documents:document_version_multiple_delete'
)
link_document_version_edit = Link(
    args='resolved_object.pk', icon=icon_document_version_edit,
    permission=permission_document_version_edit,
    text=_(message='Edit'), view='documents:document_version_edit'
)
link_document_version_list = Link(
    args='resolved_object.pk', icon=icon_document_version_list,
    permission=permission_document_version_view, text=_(message='Versions'),
    view='documents:document_version_list'
)
link_document_version_modification = Link(
    args='resolved_object.pk', icon=icon_document_version_modification,
    permission=permission_document_version_edit, text=_(message='Modify'),
    view='documents:document_version_modify'
)
link_document_version_preview = Link(
    args='resolved_object.pk', icon=icon_document_version_preview,
    permission=permission_document_version_view,
    text=_(message='Preview'), view='documents:document_version_preview'
)
link_document_version_print_form = Link(
    args='resolved_object.id', icon=icon_document_version_print,
    permission=permission_document_version_print, text=_(message='Print'),
    view='documents:document_version_print_form'
)
link_document_version_return_to_document = Link(
    args='resolved_object.document.pk',
    icon=icon_document_version_return_document,
    permission=permission_document_view, text=_(message='Document'),
    view='documents:document_preview'
)
link_document_version_return_list = Link(
    args='resolved_object.document.pk',
    icon=icon_document_version_return_list,
    permission=permission_document_version_view, text=_(message='Versions'),
    view='documents:document_version_list'
)
link_document_version_transformations_clear = Link(
    args='resolved_object.id',
    icon=icon_document_version_transformation_list_clear,
    permission=permission_transformation_delete,
    text=_(message='Clear transformations'),
    view='documents:document_version_transformations_clear'
)
link_document_version_multiple_transformations_clear = Link(
    icon=icon_document_version_transformation_list_clear,
    permission=permission_transformation_delete,
    text=_(message='Clear transformations'),
    view='documents:document_version_multiple_transformations_clear'
)
link_document_version_transformations_clone = Link(
    args='resolved_object.id',
    icon=icon_document_version_transformation_list_clone,
    permission=permission_transformation_edit,
    text=_(message='Clone transformations'),
    view='documents:document_version_transformations_clone'
)



from django.contrib import admin

from .models import DocumentFilePageContent


@admin.register(DocumentFilePageContent)
class DocumentFilePageContentAdmin(admin.ModelAdmin):
    list_display = ('document_file_page',)



from django.utils.translation import gettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_c

queue_document_exports = CeleryQueue(
    name='documents_exports', label=_(message='Documents exports'),
    worker=worker_c
)

queue_document_exports.add_task_type(
    dotted_path='mayan.apps.document_exports.tasks.task_document_version_export',
    label=_(message='Export a document version to a PDF')
)



from django.core.exceptions import ValidationError as DjangoValidationError
from django.utils.translation import gettext_lazy as _

from rest_framework.exceptions import ValidationError

from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import (
    FilteredPrimaryKeyRelatedField, MultiKwargHyperlinkedIdentityField
)

from ..models.workflow_state_escalation_models import WorkflowStateEscalation


class WorkflowTemplateStateEscalationSerializer(serializers.HyperlinkedModelSerializer):
    url = MultiKwargHyperlinkedIdentityField(
        label=_(message='URL'), view_kwargs=(
            {
                'lookup_field': 'state__workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'state_id',
                'lookup_url_kwarg': 'workflow_template_state_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'workflow_template_state_escalation_id'
            }
        ), view_name='rest_api:workflow-template-state-escalation-detail'
    )
    workflow_template_state_id = serializers.IntegerField(
        label=_(message='Workflow template state ID'), read_only=True,
        source='state_id'
    )
    workflow_template_state_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Workflow template state URL'), view_kwargs=(
            {
                'lookup_field': 'state__workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'state_id',
                'lookup_url_kwarg': 'workflow_template_state_id'
            }
        ), view_name='rest_api:workflow-template-state-detail'
    )
    workflow_template_transition_id = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the workflow template transition to be added.'
        ), label=_(message='Workflow template transition ID'),
        source='transition_id',
        source_queryset_method='get_workflow_template_transition_queryset'
    )
    workflow_template_transition_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Workflow template transition URL'), view_kwargs=(
            {
                'lookup_field': 'transition__workflow_id',
                'lookup_url_kwarg': 'workflow_template_id'
            },
            {
                'lookup_field': 'transition_id',
                'lookup_url_kwarg': 'workflow_template_transition_id'
            }
        ), view_name='rest_api:workflow-template-transition-detail'
    )

    class Meta:
        fields = (
            'amount', 'comment', 'condition', 'enabled', 'id', 'priority',
            'url', 'unit', 'workflow_template_state_id',
            'workflow_template_state_url', 'workflow_template_transition_id',
            'workflow_template_transition_url'
        )
        model = WorkflowStateEscalation
        read_only_fields = (
            'id', 'url', 'workflow_template_state_id',
            'workflow_template_state_url',
            'workflow_template_transition_url'
        )

    def create(self, validated_data):
        validated_data['transition'] = validated_data.pop(
            'transition_id'
        )

        return super().create(
            validated_data=validated_data
        )

    def get_workflow_template_transition_queryset(self):
        return self.context['workflow_template'].transitions.filter(
            origin_state=self.context['workflow_template_state']
        )

    def update(self, instance, validated_data):
        if 'transition_id' in validated_data:
            validated_data['transition'] = validated_data.pop(
                'transition_id'
            )

        return super().update(
            instance=instance, validated_data=validated_data
        )

    def validate(self, attrs):
        if self.instance:
            return attrs

        kwargs = attrs.copy()
        kwargs['state'] = self.context['workflow_template_state']
        kwargs['transition_id'] = attrs['transition_id'].pk

        instance = WorkflowStateEscalation(**kwargs)
        try:
            instance.full_clean()
        except DjangoValidationError as exception:
            raise ValidationError(detail=exception)

        return attrs



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('web_links', '0002_auto_20191210_0436')
    ]

    operations = [
        migrations.AlterField(
            model_name='weblink',
            name='template',
            field=models.TextField(
                help_text='Template that will be used to craft the '
                'final URL of the web link.', verbose_name='Template'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

from django_celery_beat.models import DAYS

TEST_INTERVAL_SCHEDULE_EVERY = 365
TEST_INTERVAL_SCHEDULE_PERIOD = DAYS

TEST_PERIODIC_TASK_NAME = 'test_periodic_task'
TEST_PERIODIC_TASK_TASK = 'mayan.apps.task_manager.tests.tasks.test_task'

TEST_QUEUE_LABEL = _(message='Test queue')
TEST_QUEUE_NAME = 'test_queue'
TEST_WORKER_NAME = 'test_worker'



from django.conf import settings
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.template import TemplateSyntaxError
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.models.document_models import Document
from mayan.apps.views.generics import FormView
from mayan.apps.views.http import URL
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from .classes import Template
from .forms import DocumentTemplateSandboxForm
from .icons import icon_template_sandbox
from .permissions import permission_template_sandbox


class DocumentTemplateSandboxView(ExternalObjectViewMixin, FormView):
    external_object_class = Document
    external_object_permission = permission_template_sandbox
    external_object_pk_url_kwarg = 'document_id'
    form_class = DocumentTemplateSandboxForm
    view_icon = icon_template_sandbox

    def form_valid(self, form):
        path = reverse(
            viewname='templating:document_template_sandbox',
            kwargs={'document_id': self.external_object.pk}
        )
        url = URL(
            path=path, query={
                'template': form.cleaned_data['template']
            }
        )

        return HttpResponseRedirect(
            redirect_to=url.to_string()
        )

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'title': _(message='Template sandbox for: %s') % self.external_object
        }

    def get_form_extra_kwargs(self):
        return {'model': Document, 'model_variable': 'document'}

    def get_initial(self):
        if settings.DEBUG:
            exception_list = (TemplateSyntaxError,)
        else:
            exception_list = (Exception, TemplateSyntaxError,)

        template_string = self.request.GET.get('template', '')
        try:
            template = Template(template_string=template_string)
            result = template.render(
                context={'document': self.external_object}
            )
        except exception_list as exception:
            result = ''
            error_message = _(
                'Template error; %(exception)s'
            ) % {
                'exception': exception
            }
            messages.error(message=error_message, request=self.request)

        return {
            'result': result, 'template': template_string
        }



from mayan.apps.common.tests.mixins import ManagementCommandTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..classes import PlatformTemplate
from ..literals import COMMAND_NAME_PLATFORM_TEMPLATE

from .literals import (
    TEST_TEMPLATE_LABEL, TEST_TEMPLATE_NAME, TEST_TEMPLATE_STRING,
    TEST_TEMPLATE_STRING_RENDER, TEST_TEMPLATE_STRING_RENDER_ALT,
    TEST_TEMPLATE_VARIABLE_VALUE, TEST_TEMPLATE_VARIABLE_VALUE_ALT
)


class TestPlatformTemplate(PlatformTemplate):
    context = {'test_template_variable': TEST_TEMPLATE_VARIABLE_VALUE}
    label = TEST_TEMPLATE_LABEL
    name = TEST_TEMPLATE_NAME
    template_string = TEST_TEMPLATE_STRING


PlatformTemplate.register(klass=TestPlatformTemplate)


class PlatformTemplateManagementCommandTestCase(
    ManagementCommandTestMixin, BaseTestCase
):
    _test_management_command_name = COMMAND_NAME_PLATFORM_TEMPLATE

    def test_platform_template_simple(self):
        stdout, stderr = self._call_test_management_command(
            TEST_TEMPLATE_NAME
        )
        self.assertEqual(stdout, TEST_TEMPLATE_STRING_RENDER)

    def test_platform_template_context(self):
        args = (
            TEST_TEMPLATE_NAME, '--context',
            'test_template_variable: {}'.format(
                TEST_TEMPLATE_VARIABLE_VALUE_ALT
            )
        )

        stdout, stderr = self._call_test_management_command(*args)
        self.assertEqual(stdout, TEST_TEMPLATE_STRING_RENDER_ALT)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('announcements', '0003_auto_20210414_2323')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='announcement',
            options={
                'ordering': ('label',), 'verbose_name': 'Announcement',
                'verbose_name_plural': 'Announcements'
            }
        )
    ]



from collections import deque
import functools
import logging

import elasticsearch
from elasticsearch import Elasticsearch, helpers
import elasticsearch_dsl
from elasticsearch_dsl import MultiSearch, Q, Search

from ..exceptions import (
    DynamicSearchBackendException, DynamicSearchValueTransformationError
)
from ..search_backends import SearchBackend
from ..search_fields import SearchFieldVirtualAllFields
from ..search_models import SearchModel
from ..search_query_types import (
    BackendQueryType, QueryTypeExact, QueryTypeFuzzy, QueryTypeGreaterThan,
    QueryTypeGreaterThanOrEqual, QueryTypeLessThan, QueryTypeLessThanOrEqual,
    QueryTypePartial, QueryTypeRange, QueryTypeRangeExclusive,
    QueryTypeRegularExpression
)

from .literals.elasticsearch_literals import (
    DEFAULT_ELASTICSEARCH_CLIENT_MAXSIZE,
    DEFAULT_ELASTICSEARCH_CLIENT_SNIFF_ON_CONNECTION_FAIL,
    DEFAULT_ELASTICSEARCH_CLIENT_SNIFF_ON_START,
    DEFAULT_ELASTICSEARCH_CLIENT_SNIFFER_TIMEOUT,
    DEFAULT_ELASTICSEARCH_HOST, DEFAULT_ELASTICSEARCH_INDICES_NAMESPACE,
    DJANGO_TO_ELASTICSEARCH_FIELD_MAP, MAXIMUM_API_ATTEMPT_COUNT
)

logger = logging.getLogger(name=__name__)


class BackendQueryTypeExact(BackendQueryType):
    query_type = QueryTypeExact

    def do_resolve(self):
        if self.value is not None:
            if self.is_quoted_value:
                template = '"{}"'
            else:
                template = '{}'

            if not self.value:
                # Empty values cannot be quoted.
                template = '{}'

                if self.is_quoted_value:
                    if self.get_search_backend_field_type() == elasticsearch_dsl.field.Text:
                        return Q(
                            'bool', must_not=(
                                Q(
                                    'wildcard',
                                    **{self.search_field.field_name: '*'}
                                )
                            )
                        )

            if self.is_quoted_value:
                return Q(
                    name_or_query='match_phrase', _expand__to_dot=False,
                    **{
                        self.search_field.field_name: template.format(self.value)
                    }
                )
            else:
                return Q(
                    name_or_query='match', _expand__to_dot=False,
                    **{
                        self.search_field.field_name: template.format(self.value)
                    }
                )


class BackendQueryFuzzy(BackendQueryType):
    query_type = QueryTypeFuzzy

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='fuzzy', _expand__to_dot=False,
                **{self.search_field.field_name: self.value}
            )


class BackendQueryTypeGreaterThan(BackendQueryType):
    query_type = QueryTypeGreaterThan

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='range', _expand__to_dot=False,
                **{
                    self.search_field.field_name: {'gt': self.value}
                }
            )


class BackendQueryTypeGreaterThanOrEqual(BackendQueryType):
    query_type = QueryTypeGreaterThanOrEqual

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='range', _expand__to_dot=False,
                **{
                    self.search_field.field_name: {'gte': self.value}
                }
            )


class BackendQueryTypeLessThan(BackendQueryType):
    query_type = QueryTypeLessThan

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='range', _expand__to_dot=False,
                **{
                    self.search_field.field_name: {'lt': self.value}
                }
            )


class BackendQueryTypeLessThanOrEqual(BackendQueryType):
    query_type = QueryTypeLessThanOrEqual

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='range', _expand__to_dot=False,
                **{
                    self.search_field.field_name: {'lte': self.value}
                }
            )


class BackendQueryTypePartial(BackendQueryType):
    query_type = QueryTypePartial

    def do_resolve(self):
        if self.value is not None:
            if self.get_search_backend_field_type() != elasticsearch_dsl.field.Date:
                if self.is_quoted_value:
                    return Q(
                        name_or_query='match_phrase', _expand__to_dot=False,
                        **{
                            self.search_field.field_name: '{}'.format(self.value)
                        }
                    )
                else:
                    if self.get_search_backend_field_type() != elasticsearch_dsl.field.Integer:
                        return Q(
                            name_or_query='wildcard', _expand__to_dot=False,
                            **{
                                self.search_field.field_name: '*{}*'.format(self.value)
                            }
                        )


class BackendQueryTypeRange(BackendQueryType):
    query_type = QueryTypeRange

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='range', _expand__to_dot=False,
                **{
                    self.search_field.field_name: {
                        'gte': self.value[0], 'lte': self.value[1]
                    }
                }
            )


class BackendQueryTypeRangeExclusive(BackendQueryType):
    query_type = QueryTypeRangeExclusive

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='range', _expand__to_dot=False,
                **{
                    self.search_field.field_name: {
                        'gt': self.value[0], 'lt': self.value[1]
                    }
                }
            )


class BackendQueryTypeRegularExpression(BackendQueryType):
    query_type = QueryTypeRegularExpression

    def do_resolve(self):
        if self.value is not None:
            return Q(
                name_or_query='regexp', _expand__to_dot=False,
                **{self.search_field.field_name: self.value}
            )


class ElasticSearchBackend(SearchBackend):
    feature_reindex = True
    field_type_mapping = DJANGO_TO_ELASTICSEARCH_FIELD_MAP

    def __init__(
        self, client_http_auth=None, client_host=DEFAULT_ELASTICSEARCH_HOST,
        client_hosts=None,
        client_maxsize=DEFAULT_ELASTICSEARCH_CLIENT_MAXSIZE,
        client_port=None, client_scheme=None,
        client_sniff_on_connection_fail=DEFAULT_ELASTICSEARCH_CLIENT_SNIFF_ON_CONNECTION_FAIL,
        client_sniff_on_start=DEFAULT_ELASTICSEARCH_CLIENT_SNIFF_ON_START,
        client_sniffer_timeout=DEFAULT_ELASTICSEARCH_CLIENT_SNIFFER_TIMEOUT,
        indices_namespace=DEFAULT_ELASTICSEARCH_INDICES_NAMESPACE,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.indices_namespace = indices_namespace

        self.client_kwargs = {
            'hosts': client_hosts or (client_host,),
            'http_auth': client_http_auth, 'maxsize': client_maxsize,
            'port': client_port, 'scheme': client_scheme,
            'sniff_on_start': client_sniff_on_start,
            'sniff_on_connection_fail': client_sniff_on_connection_fail,
            'sniffer_timeout': client_sniffer_timeout
        }

        if self._test_mode:
            self.indices_namespace = 'mayan-test'

    def _do_response_process(self, response):
        result = set()

        for hit in response:
            result.add(
                int(
                    hit['id']
                )
            )

        return result

    def do_search_execute(self, search):
        try:
            return search.execute()
        except elasticsearch.exceptions.NotFoundError as exception:
            raise DynamicSearchBackendException(
                'Index not found. Make sure the search engine '
                'was properly initialized or upgraded if '
                'it already existed.'
            ) from exception

    def _get_client(self):
        return Elasticsearch(**self.client_kwargs)

    def _get_index_name(self, search_model):
        return '{}-{}'.format(
            self.indices_namespace, search_model.model_name.lower()
        )

    @functools.cache
    def _get_search_model_index_mappings(self, search_model):
        mappings = {}

        field_map = self.get_resolved_field_type_map(
            search_model=search_model
        )
        for field_name, search_field_data in field_map.items():
            mappings[field_name] = {
                'type': search_field_data['field'].name
            }

            if 'analyzer' in search_field_data:
                mappings[field_name]['analyzer'] = search_field_data['analyzer']

        return mappings

    def _get_status(self):
        client = self._get_client()
        result = []

        title = 'Elastic Search search model indexing status'
        result.append(title)
        result.append(len(title) * '=')

        self.refresh()

        for search_model in SearchModel.all():
            index_name = self._get_index_name(search_model=search_model)
            try:
                index_stats = client.count(index=index_name)
            except elasticsearch.exceptions.NotFoundError:
                index_stats = {}

            count = index_stats.get('count', 'None')

            result.append(
                '{}: {}'.format(search_model.label, count)
            )

        return '\n'.join(result)

    def _initialize(self):
        self._update_mappings()

    def _search(
        self, limit, search_field, query_type, value, is_quoted_value=False,
        is_raw_value=False
    ):
        self.do_query_type_verify(
            query_type=query_type, search_field=search_field
        )

        client = self._get_client()
        index_name = self._get_index_name(
            search_model=search_field.search_model
        )

        if isinstance(search_field, SearchFieldVirtualAllFields):
            multi_search = MultiSearch(index=index_name, using=client)

            for search_field in search_field.field_composition:
                try:
                    search_field_query = query_type.resolve_for_backend(
                        is_quoted_value=is_quoted_value,
                        is_raw_value=is_raw_value, search_backend=self,
                        search_field=search_field, value=value
                    )
                except DynamicSearchValueTransformationError:
                    """Skip the search field."""
                else:
                    if search_field_query is not None:
                        index_name = self._get_index_name(
                            search_model=search_field.search_model
                        )

                        multi_search = multi_search.add(
                            Search().query(
                                search_field_query
                            )
                        )

            if multi_search.to_dict():
                responses = self.do_search_execute(search=multi_search)
                result = set()

                for response in responses:
                    result.update(
                        self._do_response_process(response=response)
                    )

                return result
            else:
                return ()
        else:
            search = Search(index=index_name, using=client)

            try:
                search_field_query = query_type.resolve_for_backend(
                    is_quoted_value=is_quoted_value,
                    is_raw_value=is_raw_value, search_backend=self,
                    search_field=search_field, value=value
                )
            except DynamicSearchValueTransformationError:
                return ()
            else:
                if search_field_query is None:
                    return ()
                else:
                    search = search.source(None).query(search_field_query)
                    response = self.do_search_execute(
                        search=search[0:limit]
                    )
                    return self._do_response_process(response=response)

    def _update_mappings(self, search_model=None):
        client = self._get_client()

        if search_model:
            search_models = (search_model,)
        else:
            search_models = SearchModel.all()

        for search_model in search_models:
            index_name = self._get_index_name(search_model=search_model)

            mappings = self._get_search_model_index_mappings(
                search_model=search_model
            )

            try:
                client.indices.delete(index=index_name)
            except elasticsearch.exceptions.NotFoundError:
                """
                Non fatal, might be that this is the first time
                the method is executed. Proceed.
                """

            try:
                client.indices.create(
                    index=index_name,
                    body={
                        'mappings': {
                            'properties': mappings
                        }
                    }
                )
            except elasticsearch.exceptions.RequestError:
                try:
                    client.indices.put_mapping(
                        index=index_name,
                        body={
                            'properties': mappings
                        }
                    )
                except elasticsearch.exceptions.RequestError:
                    """
                    There are mapping changes that were not allowed.
                    Example: Text to Keyword.
                    Boot up regardless and allow user to reindex to delete
                    old indices.
                    """

    def deindex_instance(self, instance):
        search_model = SearchModel.get_for_model(instance=instance)
        client = self._get_client()
        client.delete(
            id=instance.pk,
            index=self._get_index_name(search_model=search_model)
        )

    def index_instance(
        self, instance, exclude_model=None, exclude_kwargs=None
    ):
        search_model = SearchModel.get_for_model(instance=instance)

        document = search_model.populate(
            exclude_kwargs=exclude_kwargs, exclude_model=exclude_model,
            instance=instance, search_backend=self
        )
        self._get_client().index(
            index=self._get_index_name(search_model=search_model),
            id=instance.pk, document=document
        )

    def index_instances(self, search_model, id_list):
        client = self._get_client()
        index_name = self._get_index_name(search_model=search_model)

        def generate_actions():
            queryset = search_model.get_queryset()

            queryset = queryset.filter(pk__in=id_list)

            for instance in queryset:
                kwargs = search_model.populate(
                    search_backend=self, instance=instance
                )
                kwargs['_id'] = kwargs['id']

                yield kwargs

        bulk_indexing_generator = helpers.streaming_bulk(
            actions=generate_actions(), client=client, index=index_name,
            yield_ok=False
        )

        deque(iterable=bulk_indexing_generator, maxlen=0)

    def refresh(self):
        attempt_count = 0
        client = self._get_client()
        search_model_index = 0
        search_models = SearchModel.all()

        while True:
            search_model = search_models[search_model_index]
            index_name = self._get_index_name(search_model=search_model)

            try:
                client.indices.refresh(index=index_name)
            except elasticsearch.exceptions.NotFoundError as exception:
                attempt_count += 1

                if attempt_count > MAXIMUM_API_ATTEMPT_COUNT:
                    raise DynamicSearchBackendException(
                        'Refresh attempt count exceeded the maximum'
                        ' of `{}`.'.format(
                            MAXIMUM_API_ATTEMPT_COUNT
                        )
                    ) from exception
            else:
                attempt_count = 0
                search_model_index += 1
                if search_model_index >= len(search_models):
                    break

    def reset(self, search_model=None):
        self.tear_down(search_model=search_model)
        self._update_mappings(search_model=search_model)

    def tear_down(self, search_model=None):
        client = self._get_client()

        if search_model:
            search_models = (search_model,)
        else:
            search_models = SearchModel.all()

        for search_model in search_models:
            try:
                client.indices.delete(
                    index=self._get_index_name(search_model=search_model)
                )
            except elasticsearch.exceptions.NotFoundError:
                """Ignore non existent indexes."""


BackendQueryType.register(
    klass=BackendQueryTypeExact, search_backend=ElasticSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryFuzzy, search_backend=ElasticSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeGreaterThan, search_backend=ElasticSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeGreaterThanOrEqual,
    search_backend=ElasticSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeLessThan, search_backend=ElasticSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeLessThanOrEqual,
    search_backend=ElasticSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypePartial, search_backend=ElasticSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRange,
    search_backend=ElasticSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRangeExclusive,
    search_backend=ElasticSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRegularExpression,
    search_backend=ElasticSearchBackend
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import MultiKwargHyperlinkedIdentityField

from .models import SignatureCapture


class SignatureCaptureSerializer(serializers.HyperlinkedModelSerializer):
    document_url = serializers.HyperlinkedIdentityField(
        label=_(message='Document URL'), lookup_field='document_id',
        lookup_url_kwarg='document_id', view_name='rest_api:document-detail'
    )
    image_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Image URL'), view_kwargs=(
            {
                'lookup_field': 'document_id',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'signature_capture_id'
            }
        ), view_name='rest_api:signature_capture-image'
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_(message='URL'), view_kwargs=(
            {
                'lookup_field': 'document_id',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'signature_capture_id'
            }
        ), view_name='rest_api:signature_capture-detail'
    )

    class Meta:
        fields = (
            'data', 'document_id', 'document_url', 'id', 'image_url',
            'internal_name', 'text', 'url'
        )
        model = SignatureCapture
        read_only_fields = (
            'document_id', 'document_url', 'id', 'image_url', 'url'
        )



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('cabinets', '0003_auto_20200330_0851')
    ]

    operations = [
        migrations.CreateModel(
            name='CabinetSearchResult',
            fields=[],
            options={
                'verbose_name': 'Cabinet',
                'verbose_name_plural': 'Cabinets',
                'proxy': True,
                'indexes': [],
                'constraints': []
            },
            bases=('cabinets.cabinet',),
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_b

queue_metadata = CeleryQueue(
    label=_(message='Metadata'), name='metadata', worker=worker_b
)

queue_metadata.add_task_type(
    label=_(message='Remove metadata type'),
    dotted_path='mayan.apps.metadata.tasks.task_remove_metadata_type'
)
queue_metadata.add_task_type(
    label=_(message='Add required metadata type'),
    dotted_path='mayan.apps.metadata.tasks.task_add_required_metadata_type'
)



from django.db import migrations
from django.utils.text import slugify


def code_slugify_internal_name(apps, schema_editor):
    StoredCredential = apps.get_model(
        app_label='credentials', model_name='StoredCredential'
    )

    for stored_credential in StoredCredential.objects.using(alias=schema_editor.connection.alias).all():
        stored_credential.internal_name = slugify(
            value=stored_credential.label
        )
        stored_credential.save()


class Migration(migrations.Migration):
    dependencies = [
        ('credentials', '0003_auto_20210207_0823')
    ]

    operations = [
        migrations.RunPython(
            code=code_slugify_internal_name,
            reverse_code=migrations.RunPython.noop
        )
    ]



from django.http.request import HttpRequest
from django.utils.functional import cached_property

from .settings import setting_organization_installation_url


def patch_HttpRequest():
    class MockClass:
        @cached_property
        def _patched_current_scheme_host(self):
            if setting_organization_installation_url.value:
                return setting_organization_installation_url.value
            else:
                return self._original_current_scheme_host

    _original_current_scheme_host = HttpRequest._current_scheme_host

    HttpRequest._current_scheme_host = MockClass._patched_current_scheme_host
    HttpRequest._original_current_scheme_host = _original_current_scheme_host



import logging

from django.apps import apps
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.lock_manager.backends.base import LockingBackend
from mayan.apps.lock_manager.exceptions import LockError
from mayan.apps.templating.classes import Template

logger = logging.getLogger(name=__name__)


class IndexInstanceBusinessLogicMixin:
    def _delete_empty_nodes(self):
        self.index_instance_root_node.get_children().filter(
            children=None, documents=None
        ).delete()

    def _document_add(self, document, index_instance_node_parent):
        for index_template_node in index_instance_node_parent.index_template_node.get_children().filter(enabled=True):
            try:
                template = Template(
                    template_string=index_template_node.expression
                )
                result = template.render(
                    context={'document': document}
                )
            except Exception as exception:
                logger.debug('Evaluating error: %s', exception)
                error_message = _(
                    'Error indexing document: %(document)s; expression: '
                    '%(expression)s; %(exception)s'
                ) % {
                    'document': document,
                    'exception': exception,
                    'expression': index_template_node.expression
                }
                logger.debug(msg=error_message)
            else:
                logger.debug('Evaluation result: %s', result)

                if result:
                    index_instance_node, created = index_template_node.index_instance_nodes.get_or_create(
                        parent=index_instance_node_parent,
                        value=result
                    )

                    if index_template_node.link_documents:
                        index_instance_node.documents.add(document)

                    self._document_add(
                        document=document,
                        index_instance_node_parent=index_instance_node
                    )

    def document_nodes_delete(self, document):
        IndexInstanceNode = apps.get_model(
            app_label='document_indexing', model_name='IndexInstanceNode'
        )

        IndexInstanceNode.documents.through.objects.filter(
            document=document,
            indexinstancenode__index_template_node__enabled=True,
            indexinstancenode__index_template_node__index=self,
            indexinstancenode__index_template_node__index__enabled=True,
            indexinstancenode__index_template_node__index__document_types=document.document_type
        ).delete()

    def document_add(self, document):
        """
        Method to start the indexing process for a document. The entire
        process happens inside one transaction. The document is first
        removed from all the index nodes to which it already belongs.
        The different index templates that match this document's type
        are evaluated and for each result a node is fetched or created and
        the document is added to that node.
        """
        logger.debug('Index; Indexing document: %s', document)

        if Document.valid.filter(pk=document.pk).exists() and self.enabled and self.document_types.filter(pk=document.document_type.pk).exists():
            try:
                locking_backend = LockingBackend.get_backend()

                lock_index_instance = locking_backend.acquire_lock(
                    name=self.get_lock_string()
                )
            except LockError:
                raise
            else:
                try:
                    lock_document = locking_backend.acquire_lock(
                        name=self.get_document_lock_string(document=document)
                    )
                except LockError:
                    raise
                else:
                    try:
                        self.initialize_index_instance_root_node_node()

                        self.document_nodes_delete(document=document)

                        self._document_add(
                            document=document,
                            index_instance_node_parent=self.index_instance_root_node
                        )

                        self._delete_empty_nodes()
                    finally:
                        lock_document.release()
                finally:
                    lock_index_instance.release()

    def document_remove(self, document):
        if self.enabled and self.document_types.filter(pk=document.document_type.pk).exists():
            try:
                lock_index_instance = LockingBackend.get_backend().acquire_lock(
                    name=self.get_lock_string()
                )
            except LockError:
                raise
            else:
                try:
                    lock_document = LockingBackend.get_backend().acquire_lock(
                        name=self.get_document_lock_string(
                            document=document
                        )
                    )
                except LockError:
                    raise
                else:
                    try:
                        self.document_nodes_delete(document=document)
                        self._delete_empty_nodes()
                    finally:
                        lock_document.release()
                finally:
                    lock_index_instance.release()

    def get_children(self):
        return self.index_instance_root_node.get_children()

    def get_document_lock_string(self, document):
        return 'indexing:document_{}'.format(document.pk)

    def get_descendants(self):
        return self.index_instance_root_node.get_descendants()

    def get_descendants_count(self):
        return self.index_instance_root_node.get_descendants_count()

    get_descendants_count.help_text = _(
        'Total number of nodes with unique values this item contains.'
    )

    def get_descendants_document_count(self, user):
        return self.index_instance_root_node.get_descendants_document_count(
            user=user
        )

    get_descendants_document_count.help_text = _(
        'Total number of unique documents this item contains.'
    )

    def get_lock_string(self):
        return 'indexing:index_instance_{}'.format(self.pk)

    def get_level_count(self):
        return self.index_instance_root_node.get_level_count()

    get_level_count.help_text = _(
        'Total number of node levels this item contains.'
    )

    def get_root(self):
        """Compatibility method."""
        return self.index_instance_root_node

    @property
    def index_instance_root_node(self):
        return self.index_template_root_node.get_index_instance_root_node()

    def initialize_index_instance_root_node_node(self):
        return self.index_template_root_node.initialize_index_instance_root_node()


class IndexInstanceNodeBusinessLogicMixin:
    def _get_documents(self):
        return Document.valid.filter(
            pk__in=self.documents.values('pk')
        )

    def get_children_count(self):
        return self.get_children().count()

    def get_descendants_count(self):
        return self.get_descendants().count()

    get_descendants_count.help_text = IndexInstanceBusinessLogicMixin.get_descendants_count.help_text

    def get_descendants_document_count(self, user):
        queryset = Document.valid.filter(
            index_instance_nodes__in=self.get_descendants(
                include_self=True
            )
        ).distinct()

        return AccessControlList.objects.restrict_queryset(
            permission=permission_document_view,
            queryset=queryset, user=user
        ).count()

    get_descendants_document_count.help_text = IndexInstanceBusinessLogicMixin.get_descendants_document_count.help_text

    def get_documents(self, permission, user):
        """
        Provide a queryset of the documents in an index instance node.
        The queryset is filtered by access.
        """
        return AccessControlList.objects.restrict_queryset(
            permission=permission, queryset=self._get_documents(),
            user=user
        )

    def get_full_path(self):
        result = []
        for node in self.get_ancestors(include_self=True):
            if node.is_root_node():
                result.append(
                    str(
                        self.index()
                    )
                )
            else:
                result.append(
                    str(node)
                )

        return ' / '.join(result)
    get_full_path.help_text = _(
        'The path to the index including all ancestors.'
    )
    get_full_path.short_description = _(message='Full path')

    def get_level_count(self):
        return self.get_descendants().values('level').distinct().count()

    get_level_count.help_text = IndexInstanceBusinessLogicMixin.get_level_count.help_text

    def index(self):
        """
        Return's the index instance of this node instance.
        """
        IndexInstance = apps.get_model(
            app_label='document_indexing', model_name='IndexInstance'
        )

        return IndexInstance.objects.get(
            pk=self.index_template_node.index.pk
        )



import logging

from django.apps import apps
from django.db.models.signals import post_save
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import (
    menu_list_facet, menu_multi_item, menu_secondary, menu_tools
)
from mayan.apps.databases.classes import ModelFieldRelated, ModelProperty
from mayan.apps.documents.signals import signal_post_document_version_remap
from mayan.apps.events.classes import ModelEventType

from .events import (
    event_ocr_document_version_content_deleted,
    event_ocr_document_version_page_content_edited,
    event_ocr_document_version_finished,
    event_ocr_document_version_submitted
)
from .handlers import (
    handler_initialize_new_ocr_settings, handler_ocr_document_version
)
from .links import (
    link_document_version_page_ocr_content_detail_view,
    link_document_version_page_ocr_content_edit_view,
    link_document_version_ocr_content_detail,
    link_document_version_ocr_content_single_delete,
    link_document_version_ocr_content_multiple_delete,
    link_document_version_ocr_content_download,
    link_document_version_ocr_single_submit,
    link_document_version_ocr_multiple_submit,
    link_document_type_ocr_settings, link_document_type_submit
)
from .methods import (
    method_document_ocr_content, method_document_ocr_submit,
    method_document_version_ocr_content, method_document_version_ocr_submit
)
from .permissions import (
    permission_document_type_ocr_setup, permission_document_version_ocr,
    permission_document_version_ocr_content_edit,
    permission_document_version_ocr_content_view
)

logger = logging.getLogger(name=__name__)


class OCRApp(MayanAppConfig):
    app_namespace = 'ocr'
    app_url = 'ocr'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.ocr'
    verbose_name = _(message='OCR')

    def ready(self):
        super().ready()

        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        DocumentTypeOCRSettings = self.get_model(
            model_name='DocumentTypeOCRSettings'
        )
        DocumentVersion = apps.get_model(
            app_label='documents', model_name='DocumentVersion'
        )
        DocumentVersionPage = apps.get_model(
            app_label='documents', model_name='DocumentVersionPage'
        )

        Document.add_to_class(
            name='ocr_content', value=method_document_ocr_content
        )
        Document.add_to_class(
            name='submit_for_ocr', value=method_document_ocr_submit
        )
        DocumentVersion.add_to_class(
            name='ocr_content', value=method_document_version_ocr_content
        )
        DocumentVersion.add_to_class(
            name='submit_for_ocr', value=method_document_version_ocr_submit
        )

        ModelEventType.register(
            model=Document, event_types=(
                event_ocr_document_version_content_deleted,
                event_ocr_document_version_page_content_edited,
                event_ocr_document_version_finished,
                event_ocr_document_version_submitted
            )
        )
        ModelEventType.register(
            model=DocumentVersion, event_types=(
                event_ocr_document_version_content_deleted,
                event_ocr_document_version_page_content_edited,
                event_ocr_document_version_finished,
                event_ocr_document_version_submitted
            )
        )
        ModelEventType.register(
            model=DocumentVersionPage, event_types=(
                event_ocr_document_version_page_content_edited,
            )
        )

        ModelFieldRelated(
            model=Document,
            name='versions__version_pages__ocr_content__content'
        )
        ModelProperty(
            description=_(message='The OCR content.'), label='OCR content',
            model=DocumentVersionPage, name='ocr_content.content'
        )
        ModelProperty(
            description=_(
                'A generator returning the document\'s version pages OCR '
                'content.'
            ), label=_(message='OCR content'), model=Document,
            name='ocr_content'
        )

        ModelPermission.register(
            model=Document, permissions=(
                permission_document_version_ocr,
                permission_document_version_ocr_content_edit,
                permission_document_version_ocr_content_view
            )
        )
        ModelPermission.register(
            model=DocumentType, permissions=(
                permission_document_type_ocr_setup,
            )
        )
        ModelPermission.register_inheritance(
            model=DocumentTypeOCRSettings, related='document_type',
        )

        # Document type

        menu_list_facet.bind_links(
            links=(link_document_type_ocr_settings,), sources=(DocumentType,)
        )

        # Document version

        menu_list_facet.bind_links(
            links=(link_document_version_ocr_content_detail,),
            sources=(DocumentVersion,)
        )

        menu_multi_item.bind_links(
            links=(
                link_document_version_ocr_content_multiple_delete,
                link_document_version_ocr_multiple_submit
            ), sources=(DocumentVersion,)
        )

        menu_secondary.bind_links(
            links=(
                link_document_version_ocr_content_single_delete,
                link_document_version_ocr_content_download,
                link_document_version_ocr_single_submit
            ),
            sources=(
                'ocr:document_version_ocr_content_view_delete',
                'ocr:document_version_ocr_content_view',
                'ocr:document_version_ocr_content_download',
                'ocr:document_version_ocr_single_submit'
            )
        )

        # Document version page

        menu_list_facet.bind_links(
            links=(
                link_document_version_page_ocr_content_detail_view,
            ), sources=(DocumentVersionPage,)
        )

        menu_secondary.bind_links(
            links=(
                link_document_version_page_ocr_content_edit_view,
            ), sources=(
                'ocr:document_version_page_ocr_content_detail_view',
                'ocr:document_version_page_ocr_content_edit_view'
            )
        )

        menu_tools.bind_links(
            links=(
                link_document_type_submit,
            )
        )

        post_save.connect(
            dispatch_uid='ocr_handler_initialize_new_ocr_settings',
            receiver=handler_initialize_new_ocr_settings,
            sender=DocumentType
        )
        signal_post_document_version_remap.connect(
            dispatch_uid='ocr_handler_ocr_document_version',
            receiver=handler_ocr_document_version,
            sender=DocumentVersion
        )



from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.models.document_models import Document
from mayan.apps.templating.fields import ModelTemplateField
from mayan.apps.views.forms import ModelForm

from .models import WebLink


class WebLinkForm(ModelForm):
    fieldsets = (
        (
            _(message='General'), {
                'fields': ('label', 'enabled')
            },
        ),
        (
            _(message='Templating'), {
                'fields': ('template',)
            },
        ),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['template'] = ModelTemplateField(
            initial_help_text=format_lazy(
                '{} ', self.fields['template'].help_text
            ), label=self.fields['template'].label, model=Document,
            model_variable='document', required=True
        )

    class Meta:
        fields = ('label', 'enabled', 'template')
        model = WebLink



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Messaging'), name='messaging'
)

permission_message_create = namespace.add_permission(
    label=_(message='Create messages'), name='message_create'
)
permission_message_delete = namespace.add_permission(
    label=_(message='Delete messages'), name='message_delete'
)
permission_message_edit = namespace.add_permission(
    label=_(message='Edit messages'), name='message_edit'
)
permission_message_view = namespace.add_permission(
    label=_(message='View messages'), name='message_view'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .classes import LayerLink
from .icons import (
    icon_asset_create, icon_asset_delete, icon_asset_edit, icon_asset_list,
    icon_transformation_delete, icon_transformation_edit,
    icon_transformation_select
)
from .layers import layer_saved_transformations
from .permissions import (
    permission_asset_create, permission_asset_delete, permission_asset_edit,
    permission_asset_view
)
from .transformations import BaseTransformation


def conditional_active(context, resolved_link):
    return resolved_link.link.view == resolved_link.current_view_name and context['layer'] == resolved_link.link.get_layer(context=context)


def condition_valid_transformation_and_arguments(context, resolved_object):
    try:
        transformation = BaseTransformation.get(name=resolved_object.name)
    except KeyError:
        return False
    else:
        return transformation.arguments


link_asset_create = Link(
    icon=icon_asset_create, permission=permission_asset_create,
    text=_(message='Create asset'), view='converter:asset_create'
)
link_asset_multiple_delete = Link(
    icon=icon_asset_delete, tags='dangerous', text=_(message='Delete'),
    view='converter:asset_multiple_delete'
)
link_asset_single_delete = Link(
    args='object.pk', icon=icon_asset_delete,
    permission=permission_asset_delete, tags='dangerous',
    text=_(message='Delete'), view='converter:asset_single_delete'
)
link_asset_edit = Link(
    args='object.pk', icon=icon_asset_edit,
    permission=permission_asset_edit, text=_(message='Edit'),
    view='converter:asset_edit'
)
link_asset_list = Link(
    icon=icon_asset_list, text=_(message='Assets'),
    view='converter:asset_list'
)
link_asset_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='converter', model_name='Asset',
        object_permission=permission_asset_view,
        view_permission=permission_asset_create,
    ), icon=icon_asset_list, text=_(message='Assets'),
    view='converter:asset_list'
)

# Transformations

link_transformation_delete = LayerLink(
    action='delete', icon=icon_transformation_delete, tags='dangerous',
    text=_(message='Delete'), view='converter:transformation_delete'
)
link_transformation_edit = LayerLink(
    action='edit', condition=condition_valid_transformation_and_arguments,
    icon=icon_transformation_edit, text=_(message='Edit'),
    view='converter:transformation_edit'
)
link_transformation_list = LayerLink(
    action='view', conditional_active=conditional_active,
    layer=layer_saved_transformations, text=_(message='Transformations'),
    view='converter:transformation_list'
)
link_transformation_select = LayerLink(
    action='select', icon=icon_transformation_select,
    text=_(message='Select new transformation'),
    view='converter:transformation_select'
)



from ...literals import QUERY_PARAMETER_ANY_FIELD
from ...search_query_types import QueryTypeExact, QueryTypePartial
from ...settings import setting_results_limit

from .backend_mixins import BackendSearchTestMixin
from .base import SearchTestMixin, TestSearchObjectHierarchyTestMixin


class BackendSearchFieldAnyFieldTestCaseMixin:
    def test_any_field_hyphenated_value(self):
        self._test_object.label = 'P01208-06'
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value='P01208-06'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value='P01208'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value='06'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_any_field_hyphenated_value_mixed(self):
        self._test_object.label = 'P01208-06 word'
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value='P01208-06'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value='word'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_any_field_invalid_value(self):
        id_list = self._do_backend_search(
            field_name=QUERY_PARAMETER_ANY_FIELD,
            query_type=QueryTypePartial,
            value='invalid'
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object.id not in id_list)


class BackendSearchFieldDirectFieldTestCaseMixin:
    def test_direct_field_case_insensitive_search(self):
        self._test_object.label = self._test_object.label.upper()
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='label',
            query_type=QueryTypeExact,
            value=self._test_object.label.lower()
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        self._test_object.label = self._test_object.label.lower()
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='label',
            query_type=QueryTypeExact,
            value=self._test_object.label.upper()
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_direct_field_partial_search(self):
        id_list = self._do_backend_search(
            field_name='label',
            query_type=QueryTypePartial,
            value=self._test_object.label[0:4]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_direct_field_partial_case_insensitive_search(self):
        id_list = self._do_backend_search(
            field_name='label',
            query_type=QueryTypePartial,
            value=self._test_object.label.upper()[0:4]
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_direct_field_search(self):
        id_list = self._do_backend_search(
            field_name='label', query_type=QueryTypeExact,
            value=self._test_object.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_direct_field_exact_search(self):
        self._test_object.label = '123-456-789'
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='label',
            query_type=QueryTypeExact,
            value='123'
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='label', query_type=QueryTypeExact,
            value='123-456'
        )

        self.assertEqual(
            len(id_list), 0
        )

        id_list = self._do_backend_search(
            field_name='label', query_type=QueryTypeExact,
            value='123-456-789'
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

    def test_direct_field_update_search(self):
        old_label = self._test_object.label
        self._test_object.label = 'edited'
        self._test_object.save()

        id_list = self._do_backend_search(
            field_name='label', query_type=QueryTypeExact,
            value=self._test_object.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object.id in id_list)

        id_list = self._do_backend_search(
            field_name='label', query_type=QueryTypeExact,
            value=old_label
        )

        self.assertEqual(
            len(id_list), 0
        )

    def test_direct_field_delete_search(self):
        self._test_object.delete()

        id_list = self._do_backend_search(
            field_name='label', query_type=QueryTypeExact,
            value=self._test_object.label
        )

        self.assertEqual(
            len(id_list), 0
        )


class BackendLimitTestCase:
    def test_setting_search_results_limit(self):
        self._test_search_model = self._test_search_grandchild

        id_list = self._do_backend_search(
            field_name='label', query_type=QueryTypePartial,
            value='grandchild'
        )

        self.assertEqual(
            len(id_list), 2
        )
        self.assertTrue(self._test_object_grandchildren[0].id in id_list)
        self.assertTrue(self._test_object_grandchildren[1].id in id_list)

        setting_results_limit.do_value_raw_set(raw_value=1)

        id_list = self._do_backend_search(
            field_name='label', query_type=QueryTypePartial,
            value='grandchild'
        )

        self.assertEqual(
            len(id_list), 1
        )

        id_list = self._do_backend_search(
            field_name='label', limit=3,
            query_type=QueryTypePartial, value='grandchild'
        )

        self.assertEqual(
            len(id_list), 2
        )
        self.assertTrue(self._test_object_grandchildren[0].id in id_list)
        self.assertTrue(self._test_object_grandchildren[1].id in id_list)


class BackendSearchFieldManyToManyFieldTestCaseMixin:
    def test_direct_many_to_many_search(self):
        self._test_search_model = self._test_search_grandchild

        id_list = self._do_backend_search(
            field_name='attributes__label',
            query_type=QueryTypeExact,
            value=self._test_object_attribute.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object_grandchild.id in id_list)

    def test_direct_many_to_many_delete_search(self):
        self._test_search_model = self._test_search_grandchild

        self._test_object_attribute.delete()

        id_list = self._do_backend_search(
            field_name='attributes__label',
            query_type=QueryTypeExact,
            value=self._test_object_attribute.label
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandchild.id not in id_list)

    def test_direct_many_to_many_updated_search(self):
        self._test_search_model = self._test_search_grandchild

        old_label_value = self._test_object_attribute.label
        self._test_object_attribute.label = 'edited'
        self._test_object_attribute.save()

        id_list = self._do_backend_search(
            field_name='attributes__label',
            query_type=QueryTypeExact,
            value=self._test_object_attribute.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object_grandchild.id in id_list)

        id_list = self._do_backend_search(
            field_name='attributes__label',
            query_type=QueryTypeExact, value=old_label_value
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandchild.id not in id_list)

    def test_direct_many_to_many_remove_search(self):
        self._test_search_model = self._test_search_grandchild

        self._test_object_grandchild.attributes.remove(
            self._test_object_attribute
        )

        id_list = self._do_backend_search(
            field_name='attributes__label',
            query_type=QueryTypeExact,
            value=self._test_object_attribute.label
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandchild.id not in id_list)

    def test_reverse_many_to_many_search(self):
        self._test_search_model = self._test_search_attribute

        id_list = self._do_backend_search(
            field_name='children__label',
            query_type=QueryTypeExact,
            value=self._test_object_grandchild.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object_attribute.id in id_list)

    def test_reverse_many_to_many_parent_delete_search(self):
        self._test_search_model = self._test_search_attribute

        self._test_object_grandchild.delete()

        id_list = self._do_backend_search(
            field_name='children__label',
            query_type=QueryTypeExact,
            value=self._test_object_grandchild.label
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_attribute.id not in id_list)


class BackendProxyObjectTestCaseMixin:
    def test_proxy_object_field_update_search(self):
        self._test_search_model = self._test_search_grandchild

        old_label = self._test_object_grandchild_proxy.label
        self._test_object_grandchild_proxy.label = 'edited'
        self._test_object_grandchild_proxy.save()

        id_list = self._do_backend_search(
            field_name='label',
            query_type=QueryTypeExact,
            value=self._test_object_grandchild_proxy.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object_grandchild.id in id_list)

        id_list = self._do_backend_search(
            field_name='label', query_type=QueryTypeExact,
            value=old_label
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandchild.id not in id_list)

    def test_proxy_object_delete_search(self):
        self._test_search_model = self._test_search_grandchild

        id_list = self._do_backend_search(
            field_name='label', query_type=QueryTypeExact,
            value=self._test_object_grandchild.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object_grandchild.id in id_list)

        self._test_object_grandchild_proxy.delete()

        id_list = self._do_backend_search(
            field_name='label', query_type=QueryTypeExact,
            value=self._test_object_grandchild.label
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandchild.id not in id_list)

    def test_proxy_object_many_to_many_remove_search(self):
        self._test_search_model = self._test_search_grandchild

        id_list = self._do_backend_search(
            field_name='attributes__label',
            query_type=QueryTypeExact,
            value=self._test_object_attribute.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object_grandchild.id in id_list)

        self._test_object_grandchild_proxy.attributes.remove(
            self._test_object_attribute
        )

        id_list = self._do_backend_search(
            field_name='attributes__label',
            query_type=QueryTypeExact,
            value=self._test_object_attribute.label
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandchild.id not in id_list)


class BackendSearchFieldRelatedObjectDirectFieldTestCaseMixin:
    def test_related_field_search(self):
        self._test_search_model = self._test_search_grandparent

        id_list = self._do_backend_search(
            field_name='children__label',
            query_type=QueryTypeExact,
            value=self._test_object_parent.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object_grandparent.id in id_list)

    def test_related_field_delete_search(self):
        self._test_search_model = self._test_search_grandparent

        self._test_object_parent.delete()

        id_list = self._do_backend_search(
            field_name='children__label',
            query_type=QueryTypeExact,
            value=self._test_object_parent.label
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandparent.id not in id_list)

    def test_related_field_update_search(self):
        self._test_search_model = self._test_search_grandparent

        old_label_value = self._test_object_parent.label
        self._test_object_parent.label = 'edited'
        self._test_object_parent.save()

        id_list = self._do_backend_search(
            field_name='children__label',
            query_type=QueryTypeExact,
            value=self._test_object_parent.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object_grandparent.id in id_list)

        id_list = self._do_backend_search(
            field_name='children__label',
            query_type=QueryTypeExact,
            value=old_label_value
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandparent.id not in id_list)

    def test_related_field_multiple_level_search(self):
        self._test_search_model = self._test_search_grandparent

        id_list = self._do_backend_search(
            field_name='children__children__label',
            query_type=QueryTypeExact,
            value=self._test_object_grandchild.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object_grandparent.id in id_list)

    def test_related_field_multiple_level_delete_search(self):
        self._test_search_model = self._test_search_grandparent

        self._test_object_grandchild.delete()

        id_list = self._do_backend_search(
            field_name='children__children__label',
            query_type=QueryTypeExact,
            value=self._test_object_grandchild.label
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandparent.id not in id_list)

    def test_related_field_multiple_level_update_search(self):
        self._test_search_model = self._test_search_grandparent

        old_label_value = self._test_object_grandchild.label
        self._test_object_grandchild.label = 'edited'
        self._test_object_grandchild.save()

        id_list = self._do_backend_search(
            field_name='children__children__label',
            query_type=QueryTypeExact,
            value=self._test_object_grandchild.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object_grandparent.id in id_list)

        id_list = self._do_backend_search(
            field_name='children__children__label',
            query_type=QueryTypeExact,
            value=old_label_value
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandparent.id not in id_list)


class BackendSearchFieldRelatedObjectManyToManyFieldTestCaseMixin:
    def test_related_field_multiple_level_many_to_many_search(self):
        self._test_search_model = self._test_search_grandparent

        id_list = self._do_backend_search(
            field_name='children__children__attributes__label',
            query_type=QueryTypeExact,
            value=self._test_object_attribute.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object_grandparent.id in id_list)

    def test_related_field_multiple_level_many_to_many_delete_search(self):
        self._test_search_model = self._test_search_grandparent

        self._test_object_attribute.delete()

        id_list = self._do_backend_search(
            field_name='children__children__attributes__label',
            query_type=QueryTypeExact,
            value=self._test_object_attribute.label
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandparent.id not in id_list)

    def test_related_field_multiple_level_many_to_many_updated_search(self):
        self._test_search_model = self._test_search_grandparent

        old_label_value = self._test_object_attribute.label
        self._test_object_attribute.label = 'edited'
        self._test_object_attribute.save()

        id_list = self._do_backend_search(
            field_name='children__children__attributes__label',
            query_type=QueryTypeExact,
            value=self._test_object_attribute.label
        )

        self.assertEqual(
            len(id_list), 1
        )
        self.assertTrue(self._test_object_grandparent.id in id_list)

        id_list = self._do_backend_search(
            field_name='children__children__attributes__label',
            query_type=QueryTypeExact,
            value=old_label_value
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandparent.id not in id_list)

    def test_related_field_multiple_level_many_to_many_remove_search(self):
        self._test_search_model = self._test_search_grandparent

        self._test_object_grandchild.attributes.remove(
            self._test_object_attribute
        )

        id_list = self._do_backend_search(
            field_name='children__children__attributes__label',
            query_type=QueryTypeExact,
            value=self._test_object_attribute.label
        )

        self.assertEqual(
            len(id_list), 0
        )
        self.assertTrue(self._test_object_grandparent.id not in id_list)


class BackendSearchFieldTestCaseMixin(
    BackendLimitTestCase, BackendProxyObjectTestCaseMixin,
    BackendSearchFieldAnyFieldTestCaseMixin,
    BackendSearchFieldDirectFieldTestCaseMixin,
    BackendSearchFieldManyToManyFieldTestCaseMixin,
    BackendSearchFieldRelatedObjectDirectFieldTestCaseMixin,
    BackendSearchFieldRelatedObjectManyToManyFieldTestCaseMixin,
    BackendSearchTestMixin, TestSearchObjectHierarchyTestMixin,
    SearchTestMixin
):
    """
    Consolidated backend test case mixin.
    """



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0086_populate_latest_file')
    ]

    operations = [
        migrations.AddField(
            model_name='document', name='version_active',
            field=models.OneToOneField(
                blank=True, null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='document_active',
                to='documents.documentversion',
                verbose_name='Active document version'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.settings.literals import (
    MAYAN_WORKER_A_CONCURRENCY, MAYAN_WORKER_A_MAX_MEMORY_PER_CHILD,
    MAYAN_WORKER_A_MAX_TASKS_PER_CHILD, MAYAN_WORKER_A_NICE_LEVEL,
    MAYAN_WORKER_B_CONCURRENCY, MAYAN_WORKER_B_MAX_MEMORY_PER_CHILD,
    MAYAN_WORKER_B_MAX_TASKS_PER_CHILD, MAYAN_WORKER_B_NICE_LEVEL,
    MAYAN_WORKER_C_CONCURRENCY, MAYAN_WORKER_C_MAX_MEMORY_PER_CHILD,
    MAYAN_WORKER_C_MAX_TASKS_PER_CHILD, MAYAN_WORKER_C_NICE_LEVEL,
    MAYAN_WORKER_D_CONCURRENCY, MAYAN_WORKER_D_MAX_MEMORY_PER_CHILD,
    MAYAN_WORKER_D_MAX_TASKS_PER_CHILD, MAYAN_WORKER_D_NICE_LEVEL,
    MAYAN_WORKER_E_CONCURRENCY, MAYAN_WORKER_E_MAX_MEMORY_PER_CHILD,
    MAYAN_WORKER_E_MAX_TASKS_PER_CHILD, MAYAN_WORKER_E_NICE_LEVEL
)

from .classes import Worker

worker_a = Worker(
    concurrency=MAYAN_WORKER_A_CONCURRENCY,
    description=_(message='Low latency high volume tasks'),
    label=('Worker A'),
    maximum_memory_per_child=MAYAN_WORKER_A_MAX_MEMORY_PER_CHILD,
    maximum_tasks_per_child=MAYAN_WORKER_A_MAX_TASKS_PER_CHILD,
    name='worker_a', nice_level=MAYAN_WORKER_A_NICE_LEVEL
)
worker_b = Worker(
    concurrency=MAYAN_WORKER_B_CONCURRENCY,
    description=_(message='Medium latency tasks'), label=('Worker B'),
    maximum_memory_per_child=MAYAN_WORKER_B_MAX_MEMORY_PER_CHILD,
    maximum_tasks_per_child=MAYAN_WORKER_B_MAX_TASKS_PER_CHILD,
    name='worker_b', nice_level=MAYAN_WORKER_B_NICE_LEVEL
)
worker_c = Worker(
    concurrency=MAYAN_WORKER_C_CONCURRENCY,
    description=_(message='Medium latency tasks'), label=('Worker C'),
    maximum_memory_per_child=MAYAN_WORKER_C_MAX_MEMORY_PER_CHILD,
    maximum_tasks_per_child=MAYAN_WORKER_C_MAX_TASKS_PER_CHILD,
    name='worker_c', nice_level=MAYAN_WORKER_C_NICE_LEVEL
)
worker_d = Worker(
    concurrency=MAYAN_WORKER_D_CONCURRENCY,
    description=_(message='Low latency, long lived tasks'),
    label=('Worker D'),
    maximum_memory_per_child=MAYAN_WORKER_D_MAX_MEMORY_PER_CHILD,
    maximum_tasks_per_child=MAYAN_WORKER_D_MAX_TASKS_PER_CHILD,
    name='worker_d', nice_level=MAYAN_WORKER_D_NICE_LEVEL
)
worker_e = Worker(
    concurrency=MAYAN_WORKER_E_CONCURRENCY,
    description=_(message='Low latency, long lived tasks'),
    label=('Worker E'),
    maximum_memory_per_child=MAYAN_WORKER_E_MAX_MEMORY_PER_CHILD,
    maximum_tasks_per_child=MAYAN_WORKER_E_MAX_TASKS_PER_CHILD,
    name='worker_e', nice_level=MAYAN_WORKER_E_NICE_LEVEL
)



from django.apps import apps
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import menu_multi_item, menu_object
from mayan.apps.events.classes import ModelEventType

from .events import event_document_file_downloaded
from .links import (
    link_document_download_multiple, link_document_download_single,
    link_document_file_download_quick
)
from .methods import method_document_get_download_file_object
from .permissions import permission_document_file_download


class DocumentDownloadsApp(MayanAppConfig):
    app_namespace = 'document_downloads'
    app_url = 'document_downloads'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.document_downloads'
    verbose_name = _(message='Document downloads')

    def ready(self):
        super().ready()

        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        DocumentFile.add_to_class(
            name='get_download_file_object',
            value=method_document_get_download_file_object
        )

        ModelEventType.register(
            model=DocumentFile, event_types=(
                event_document_file_downloaded,
            )
        )

        ModelPermission.register(
            model=DocumentFile, permissions=(
                permission_document_file_download,
            )
        )

        menu_object.bind_links(
            links=(link_document_download_single,),
            sources=(Document,)
        )
        menu_object.bind_links(
            links=(link_document_file_download_quick,),
            sources=(DocumentFile,)
        )
        menu_multi_item.bind_links(
            links=(link_document_download_multiple,), sources=(Document,)
        )



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.http import HttpResponseRedirect
from django.template import RequestContext
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import (
    AddRemoveView, MultipleObjectDeleteView, SingleObjectCreateView,
    SingleObjectDetailView, SingleObjectEditView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..icons import (
    icon_current_user_detail, icon_user_create, icon_user_edit,
    icon_user_group_list, icon_user_list, icon_user_set_options,
    icon_user_setup, icon_user_single_delete
)
from ..links import link_user_create
from ..literals import FIELDS_ALL, FIELDSETS_ALL, FIELDSETS_USER
from ..permissions import (
    permission_group_edit, permission_user_create, permission_user_delete,
    permission_user_edit, permission_user_view
)
from ..querysets import get_user_queryset

from .view_mixins import DynamicUserFormFieldViewMixin


class UserCreateView(SingleObjectCreateView):
    extra_context = {
        'title': _(message='Create new user')
    }
    fields = FIELDS_ALL
    fieldsets = FIELDSETS_ALL
    model = get_user_model()
    view_icon = icon_user_create
    view_permission = permission_user_create

    def form_valid(self, form):
        super().form_valid(form=form)
        return HttpResponseRedirect(
            redirect_to=reverse(
                viewname='authentication:user_set_password', kwargs={
                    'user_id': self.object.pk
                }
            )
        )

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class UserDeleteView(MultipleObjectDeleteView):
    error_message = _(message='Error deleting user "%(instance)s"; %(exception)s')
    object_permission = permission_user_delete
    pk_url_kwarg = 'user_id'
    post_action_redirect = reverse_lazy(
        viewname='user_management:user_list'
    )
    success_message_plural = _(message='%(count)d users deleted successfully.')
    success_message_single = _(message='User "%(object)s" deleted successfully.')
    success_message_singular = _(message='%(count)d user deleted successfully.')
    title_plural = _(message='Delete the %(count)d selected users.')
    title_single = _(message='Delete user: %(object)s.')
    title_singular = _(message='Delete the %(count)d selected user.')
    view_icon = icon_user_single_delete

    def get_extra_context(self, **kwargs):
        if self.request.user in self.object_list:
            return {
                'message': _(
                    'Warning! You are about to delete your own user '
                    'account. You will lose access to the system. This '
                    'process is not reversible.'
                )
            }
        else:
            return {}

    def get_source_queryset(self):
        return get_user_queryset(user=self.request.user)


class UserDetailView(DynamicUserFormFieldViewMixin, SingleObjectDetailView):
    object_permission = permission_user_view
    pk_url_kwarg = 'user_id'
    view_icon = icon_current_user_detail

    def get_extra_context(self, **kwargs):
        return {
            'object': self.object,
            'title': _(message='Details of user: %s') % self.object
        }

    def get_source_queryset(self):
        return get_user_queryset(user=self.request.user)


class UserEditView(DynamicUserFormFieldViewMixin, SingleObjectEditView):
    fieldsets = FIELDSETS_USER
    object_permission = permission_user_edit
    pk_url_kwarg = 'user_id'
    post_action_redirect = reverse_lazy(
        viewname='user_management:user_list'
    )
    view_icon = icon_user_edit

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(message='Edit user: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_source_queryset(self):
        return get_user_queryset(user=self.request.user)


class UserGroupAddRemoveView(AddRemoveView):
    list_available_title = _(message='Available groups')
    # Translators: "User groups" here refer to the list of groups of a
    # specific user. The user's group membership.
    list_added_title = _(message='User groups')
    main_object_method_add_name = 'groups_add'
    main_object_method_remove_name = 'groups_remove'
    main_object_permission = permission_user_edit
    main_object_pk_url_kwarg = 'user_id'
    secondary_object_model = Group
    secondary_object_permission = permission_group_edit
    view_icon = icon_user_group_list

    def get_actions_extra_kwargs(self):
        return {'user': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'title': _(message='Groups of user: %s') % self.main_object
        }

    def get_list_added_queryset(self):
        return self.main_object.get_groups(
            permission=permission_group_edit, user=self.request.user
        )

    def get_main_object_source_queryset(self):
        return get_user_queryset(user=self.request.user)


class UserListView(SingleObjectListView):
    object_permission = permission_user_view
    view_icon = icon_user_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_user_setup,
            'no_results_main_link': link_user_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'User accounts can be create from this view. After creating '
                'a user account you will prompted to set a password for it. '
            ),
            'no_results_title': _(message='There are no user accounts'),
            'title': _(message='Users')
        }

    def get_source_queryset(self):
        return get_user_queryset(user=self.request.user)


class UserOptionsEditView(ExternalObjectViewMixin, SingleObjectEditView):
    external_object_permission = permission_user_edit
    external_object_pk_url_kwarg = 'user_id'
    fields = ('block_password_change',)
    view_icon = icon_user_set_options

    def get_external_object_queryset(self):
        return get_user_queryset(user=self.request.user)

    def get_extra_context(self):
        return {
            'title': _(
                'Edit options for user: %s'
            ) % self.external_object,
            'object': self.external_object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_object(self):
        return self.external_object.user_options

    def get_post_action_redirect(self):
        return reverse(viewname='user_management:user_list')



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0063_auto_20201012_0320')
    ]

    operations = [
        migrations.CreateModel(
            name='DocumentVersionPageResult',
            fields=[],
            options={
                'verbose_name': 'Document version page',
                'verbose_name_plural': 'Document version pages',
                'ordering': ('document_version__document', 'page_number'),
                'proxy': True,
                'indexes': [],
                'constraints': []
            },
            bases=('documents.documentversionpage',)
        ),
        migrations.AlterField(
            model_name='documentversionpage',
            name='document_version',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='version_pages', to='documents.DocumentVersion',
                verbose_name='Document version'
            )
        )
    ]



from django.utils.module_loading import import_string

from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin

from ..classes import FileMetadataDriver

from .literals import (
    TEST_DRIVER_CLASS_PATH, TEST_FILE_METADATA_KEY, TEST_FILE_METADATA_VALUE
)


class FileMetadataTestMixin:
    _test_document_file_metadata_driver_path = TEST_DRIVER_CLASS_PATH

    def setUp(self):
        super().setUp()
        FileMetadataDriver.load_modules()

        FileMetadataDriver.collection.do_driver_disable_all()

        if self._test_document_file_metadata_driver_path:
            self._test_document_file_metadata_driver = import_string(
                dotted_path=self._test_document_file_metadata_driver_path
            )
            self._test_document_file_metadata_driver.do_model_instance_populate()
            FileMetadataDriver.collection.do_driver_enable(
                driver=self._test_document_file_metadata_driver
            )


class DocumentFileMetadataTestMixin(
    DocumentTestMixin, FileMetadataTestMixin
):
    _test_document_file_metadata_create_auto = False

    def setUp(self):
        super().setUp()

        if self._test_document_file_metadata_create_auto:
            self._test_document_file_metadata_create()

    def _test_document_file_metadata_create(self):
        self._test_document_file_driver_entry, created = self._test_document_file.file_metadata_drivers.get_or_create(
            driver=self._test_document_file_metadata_driver.model_instance
        )

        self._test_document_file_metadata = self._test_document_file_driver_entry.entries.create(
            key=TEST_FILE_METADATA_KEY,
            value=TEST_FILE_METADATA_VALUE
        )

        self._test_document_file_metadata_path = '{}__{}'.format(
            self._test_document_file_driver_entry.driver.internal_name,
            self._test_document_file_metadata.key
        )


class DocumentFileMetadataViewTestMixin(DocumentFileMetadataTestMixin):
    def _request_document_file_metadata_driver_list_view(self):
        return self.get(
            viewname='file_metadata:document_file_metadata_driver_list',
            kwargs={'document_file_id': self._test_document_file.pk}
        )

    def _request_document_file_metadata_list_view(self):
        return self.get(
            viewname='file_metadata:document_file_metadata_driver_attribute_list',
            kwargs={
                'document_file_driver_id': self._test_document_file_driver_entry.pk
            }
        )

    def _request_document_file_metadata_single_submit_view(self):
        return self.post(
            viewname='file_metadata:document_file_metadata_single_submit',
            kwargs={'document_file_id': self._test_document_file.pk}
        )

    def _request_document_file_multiple_submit_view(self):
        return self.post(
            viewname='file_metadata:document_file_metadata_multiple_submit',
            data={
                'id_list': self._test_document_file.pk
            }
        )


class DocumentTypeViewTestMixin(FileMetadataTestMixin):
    def _request_document_type_file_metadata_settings_view(self):
        return self.get(
            viewname='file_metadata:document_type_file_metadata_settings',
            kwargs={'document_type_id': self._test_document.document_type.pk}
        )

    def _request_document_type_file_metadata_submit_view(self):
        return self.post(
            viewname='file_metadata:document_type_file_metadata_submit', data={
                'document_type': self._test_document_type.pk,
            }
        )


class FileMetadataDriverTestMixin(FileMetadataTestMixin):
    def _request_file_metadata_driver_list_view(self):
        return self.get(viewname='file_metadata:file_metadata_driver_list')



TEST_TEMPLATE_LABEL = 'test template label'
TEST_TEMPLATE_NAME = 'test_template_name'
TEST_TEMPLATE_STRING = '''
test template string
test template variable: {{ test_template_variable }}
'''
TEST_TEMPLATE_VARIABLE_VALUE = 'test_variable_value'
TEST_TEMPLATE_VARIABLE_VALUE_ALT = 'test_variable_value_alt'
TEST_TEMPLATE_STRING_RENDER = '''
test template string
test template variable: {}
'''.format(TEST_TEMPLATE_VARIABLE_VALUE)
TEST_TEMPLATE_STRING_RENDER_ALT = '''
test template string
test template variable: {}
'''.format(TEST_TEMPLATE_VARIABLE_VALUE_ALT)



from .literals import (
    TEST_DEPENDENCY_GROUP_NAME, TEST_DEPENDENCY_GROUP_ENTRY_NAME
)


class CheckVersionViewTestMixin:
    def _request_check_version_view(self):
        return self.get(viewname='dependencies:check_version_view')


class DependencyGroupEntryDetailViewTestMixin:
    def request_test_dependency_group_entry_detail_view(self):
        return self.get(
            viewname='dependencies:dependency_group_entry_detail', kwargs={
                'dependency_group_name': TEST_DEPENDENCY_GROUP_NAME,
                'dependency_group_entry_name': TEST_DEPENDENCY_GROUP_ENTRY_NAME
            }
        )



from django.core import management

from ...settings import setting_cluster


class Command(management.BaseCommand):
    help = 'Save the current settings into the configuration file.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--filepath', action='store', dest='filepath',
            help='Filename and path where to save the configuration file.'
        )

    def handle(self, *args, **options):
        filepath = options.get('filepath')
        setting_cluster.do_configuration_file_save(path=filepath)



from mayan.apps.documents.permissions import (
    permission_document_create, permission_document_file_new
)
from mayan.apps.source_compressed.source_backend_actions.mixins import SourceBackendActionMixinCompressedInteractive
from mayan.apps.source_interactive.source_backend_actions.callback_mixins import (
    SourceBackendActionMixinCallbackDocumentFileUploadInteractive,
    SourceBackendActionMixinCallbackDocumentUploadInteractive
)
from mayan.apps.sources.source_backend_actions.base import SourceBackendAction
from mayan.apps.sources.source_backend_actions.interfaces import SourceBackendActionInterfaceRequestRESTAPI
from mayan.apps.sources.source_backend_actions.mixins.document_file_mixins import SourceBackendActionMixinDocumentFileUploadInteractive
from mayan.apps.sources.source_backend_actions.mixins.document_mixins import (
    SourceBackendActionMixinDocumentInteractive,
    SourceBackendActionMixinDocumentUploadInteractive
)
from mayan.apps.sources.source_backend_actions.mixins.document_type_mixins import SourceBackendActionMixinDocumentTypeInteractive
from mayan.apps.sources.source_backend_actions.mixins.immediate_mode_mixins import SourceBackendActionMixinImmediateMode

from .arguments import argument_encoded_filename
from .mixins import (
    SourceBackendActionMixinFileStoredDeleteBase,
    SourceBackendActionMixinFileStoredDeleteInteractive,
    SourceBackendActionMixinFileStoredImage,
    SourceBackendActionMixinFileStoredInteractive,
    SourceBackendActionMixinFileStoredList
)


class SourceBackendActionFileStoredDeleteInteractive(
    SourceBackendActionMixinFileStoredDeleteInteractive, SourceBackendAction
):
    confirmation = True
    name = 'file_delete'
    permission = permission_document_create


class SourceBackendActionFileStoredDeleteInteractiveNot(
    SourceBackendActionMixinFileStoredDeleteBase, SourceBackendAction
):
    confirmation = True
    name = 'file_delete'
    permission = permission_document_create


class SourceBackendActionFileStoredDocumentFileUpload(
    SourceBackendActionMixinDocumentFileUploadInteractive,
    SourceBackendActionMixinDocumentInteractive,
    SourceBackendActionMixinFileStoredInteractive,
    SourceBackendActionMixinCallbackDocumentFileUploadInteractive,
    SourceBackendAction
):
    confirmation = True
    name = 'document_file_upload'
    permission = permission_document_file_new
    stored_file_identifier_name = 'encoded_filename'
    stored_method_name_file_cleanup = 'action_file_delete'

    class Interface:
        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            class Argument:
                encoded_filename = argument_encoded_filename


class SourceBackendActionFileStoredDocumentUpload(
    SourceBackendActionMixinDocumentUploadInteractive,
    SourceBackendActionMixinCompressedInteractive,
    SourceBackendActionMixinImmediateMode,
    SourceBackendActionMixinCallbackDocumentUploadInteractive,
    SourceBackendActionMixinDocumentTypeInteractive,
    SourceBackendActionMixinFileStoredInteractive,
    SourceBackendAction
):
    confirmation = True
    name = 'document_upload'
    permission = permission_document_create
    stored_file_identifier_name = 'encoded_filename'
    stored_method_name_file_cleanup = 'action_file_delete'

    class Interface:
        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            class Argument:
                encoded_filename = argument_encoded_filename


class SourceBackendActionFileStoredImage(
    SourceBackendActionMixinFileStoredImage, SourceBackendAction
):
    confirmation = False
    name = 'file_image'
    permission = permission_document_create


class SourceBackendActionFileStoredList(
    SourceBackendActionMixinFileStoredList, SourceBackendAction
):
    confirmation = False
    name = 'file_list'
    permission = permission_document_create



from django.apps import apps
from django.utils.translation import gettext_lazy as _

from .tasks import (
    task_index_instance_document_add, task_index_instance_document_remove
)


def handler_create_default_document_index(sender, **kwargs):
    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )
    IndexTemplate = apps.get_model(
        app_label='document_indexing', model_name='IndexTemplate'
    )

    index = IndexTemplate.objects.create(
        label=_(message='Creation date'), slug='creation_date'
    )
    for document_type in DocumentType.objects.all():
        index.document_types.add(document_type)

    root_template_node = index.index_template_root_node
    node = root_template_node.get_children().create(
        expression='{{ document.datetime_created|date:"Y" }}', index=index,
        parent=root_template_node
    )
    node.get_children().create(
        expression='{{ document.datetime_created|date:"m" }}',
        index=index, link_documents=True, parent=node
    )


def handler_event_trigger(sender, **kwargs):
    action = kwargs['instance']

    Document = apps.get_model(
        app_label='documents', model_name='Document'
    )
    IndexInstance = apps.get_model(
        app_label='document_indexing', model_name='IndexInstance'
    )

    if isinstance(action.target, Document):
        document = action.target
    elif isinstance(action.action_object, Document):
        document = action.action_object
    else:
        document = None

    if document:
        index_instance_queryset = IndexInstance.objects.filter(
            document_types=document.document_type,
            event_triggers__stored_event_type__name=kwargs['instance'].verb
        )

        for index_instance in index_instance_queryset:
            task_index_instance_document_add.apply_async(
                kwargs={
                    'document_id': document.pk,
                    'index_instance_id': index_instance.pk
                }
            )


def handler_index_document(sender, **kwargs):
    if kwargs.get('created', False):
        task_index_instance_document_add.apply_async(
            kwargs={'document_id': kwargs['instance'].pk}
        )


def handler_remove_document(sender, **kwargs):
    task_index_instance_document_remove.apply_async(
        kwargs={
            'document_id': kwargs['instance'].pk
        }
    )



from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from .events import event_message_created, event_message_edited
from .model_mixins import MessageBusinessLogicMixin


class Message(ExtraDataModelMixin, MessageBusinessLogicMixin, models.Model):
    sender_content_type = models.ForeignKey(
        blank=True, null=True, on_delete=models.CASCADE, to=ContentType,
        verbose_name=_(message='Sender content type')
    )
    sender_object_id = models.PositiveIntegerField(
        blank=True, null=True, verbose_name=_(message='Sender object ID')
    )
    sender_object = GenericForeignKey(
        ct_field='sender_content_type', fk_field='sender_object_id'
    )
    user = models.ForeignKey(
        db_index=True, on_delete=models.CASCADE, related_name='messages',
        to=settings.AUTH_USER_MODEL, verbose_name=_(message='User')
    )
    subject = models.CharField(
        max_length=255, help_text=_(message='Short description of this message.'),
        verbose_name=_(message='Subject')
    )
    body = models.TextField(
        help_text=_(message='The actual content of the message.'),
        verbose_name=_(message='Body')
    )
    read = models.BooleanField(
        default=False, help_text=_(
            'This field determines if the message has been read or not.'
        ), verbose_name=_(message='Read')
    )
    date_time = models.DateTimeField(
        auto_now_add=True, help_text=_(
            'Date and time of the message creation.'
        ), verbose_name=_(message='Creation date and time')
    )

    class Meta:
        ordering = ('-date_time',)
        verbose_name = _(message='Message')
        verbose_name_plural = _(message='Messages')

    def __str__(self):
        return self.get_label()

    def get_absolute_url(self):
        return reverse(
            viewname='messaging:message_detail', kwargs={
                'message_id': self.pk
            }
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_message_created,
            'target': 'self'
        },
        edited={
            'event': event_message_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import icon_dashboard_detail, icon_dashboard_list

link_dashboard_detail = Link(
    args='object.name', icon=icon_dashboard_detail,
    text=_(message='Details'), view='dashboards:dashboard_detail'
)
link_dashboard_list = Link(
    icon=icon_dashboard_list, text=_(message='Dashboards'),
    view='dashboards:dashboard_list'
)



from django.http import Http404

from mayan.apps.acls.models import AccessControlList

from ..classes import EventType
from ..literals import TEXT_UNKNOWN_EVENT_ID
from ..permissions import permission_events_view


class NotificationViewMixin:
    def get_source_queryset(self):
        return AccessControlList.objects.restrict_queryset(
            permission=permission_events_view,
            queryset=self.request.user.notifications.all(),
            user=self.request.user
        )


class VerbEventViewMixin:
    def dispatch(self, request, *args, **kwargs):
        self.event_type = self.get_event_type()
        return super().dispatch(request=request, *args, **kwargs)

    def get_event_type(self):
        event_type_id = self.kwargs['verb']

        try:
            return EventType.get(id=event_type_id)
        except KeyError:
            raise Http404(TEXT_UNKNOWN_EVENT_ID % event_type_id)



import base64
import hashlib

from django.utils.encoding import force_str


class HashAlgorithm:
    DEFAULT_BLOCK_SIZE = 65535
    _registry = {}
    hash_factory = None

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def register(cls, algorithm_class):
        cls._registry[algorithm_class.name] = algorithm_class

    def __init__(self, file_object, block_size=None):
        self.block_size = block_size or self.DEFAULT_BLOCK_SIZE
        self.file_object = file_object
        self.hash_object = self.hash_factory()

    def calculate(self):
        while (True):
            data = self.file_object.read(self.block_size)
            if not data:
                break

            self.hash_object.update(data)

    def get_digest(self):
        return force_str(
            s=self._get_digest()
        )


class SHA1Algorithm(HashAlgorithm):
    hash_factory = hashlib.sha1
    name = 'sha1'

    def _get_digest(self):
        return self.hash_object.hexdigest()


class SHA256Algorithm(HashAlgorithm):
    hash_factory = hashlib.sha256
    name = 'sha256'

    def _get_digest(self):
        return base64.b64encode(
            s=self.hash_object.digest()
        )


class SHA512Algorithm(SHA256Algorithm):
    hash_factory = hashlib.sha512
    name = 'sha512'


HashAlgorithm.register(algorithm_class=SHA1Algorithm)
HashAlgorithm.register(algorithm_class=SHA256Algorithm)
HashAlgorithm.register(algorithm_class=SHA512Algorithm)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0015_auto_20190701_1311')
    ]

    operations = [
        migrations.AlterField(
            model_name='workflowstate',
            name='completion',
            field=models.IntegerField(
                blank=True, default=0,
                help_text='The percent of completion that this state '
                'represents in relation to the workflow. Use numbers '
                'without the percent sign.', verbose_name='Completion'
            )
        ),
        migrations.AlterField(
            model_name='workflowstate',
            name='initial',
            field=models.BooleanField(
                default=False, help_text='The state at which the '
                'workflow will start in. Only one state can be the '
                'initial state.', verbose_name='Initial'
            )
        ),
        migrations.AlterField(
            model_name='workflowstateaction',
            name='when',
            field=models.PositiveIntegerField(
                choices=[
                    (1, 'On entry'), (2, 'On exit')
                ], default=1,
                help_text='At which moment of the state this action '
                'will execute.', verbose_name='When'
            )
        )
    ]



from django.core.management.base import BaseCommand

from ...models import UserOTPData


class Command(BaseCommand):
    help = 'Disable OTP for a user.'
    missing_args_message = 'You must provide a username.'

    def add_arguments(self, parser):
        parser.add_argument(
            dest='username', help='Username for which OTP will be disabled.',
            metavar='<username>'
        )

    def handle(self, username, **options):
        try:
            otp_data = UserOTPData.objects.get(user__username=username)
        except UserOTPData.DoesNotExist:
            result_text = 'Unknown username or OTP data is not initialized correctly.'
        else:
            if otp_data.is_enabled():
                otp_data.disable()

                if otp_data.is_enabled():
                    result_text = 'Unable to disable OTP.'
                else:
                    result_text = 'OTP disabled.'
            else:
                result_text = 'OTP is already disabled.'

        self.stdout.write(result_text)



MAILER_BACKEND_DJANGO_FILE_PATH = 'mayan.apps.mailer.mailers.DjangoFileBased'
MAILER_BACKEND_DJANGO_SMTP_PATH = 'mayan.apps.mailer.mailers.DjangoSMTP'

MAILER_BACKEND_TEST_PATH = 'mayan.apps.mailer.tests.mailers.TestBackend'

TEST_DJANGO_FILE_PATH = '/tmp/email'

TEST_EMAIL_ADDRESS = 'test@example.com'
TEST_EMAIL_BODY = 'test body'
TEST_EMAIL_BODY_HTML = '<strong>test body</strong>'
TEST_EMAIL_FROM_ADDRESS = 'from.test@example.com'
TEST_EMAIL_SUBJECT = 'test subject'

TEST_RECIPIENTS_MULTIPLE_COMMA = 'test@example.com,test2@example.com'
TEST_RECIPIENTS_MULTIPLE_COMMA_RESULT = [
    'test@example.com', 'test2@example.com'
]
TEST_RECIPIENTS_MULTIPLE_SEMICOLON = 'test@example.com;test2@example.com'
TEST_RECIPIENTS_MULTIPLE_SEMICOLON_RESULT = [
    'test@example.com', 'test2@example.com'
]
TEST_RECIPIENTS_MULTIPLE_MIXED = 'test@example.com,test2@example.com;test2@example.com'
TEST_RECIPIENTS_MULTIPLE_MIXED_RESULT = [
    'test@example.com', 'test2@example.com', 'test2@example.com'
]

TEST_MAILING_OBJECT_CONTENT = b'test_content'
TEST_MAILING_OBJECT_MIME_TYPE = 'test/mime_type'

TEST_MAILING_PROFILE_BACKEND_PATH = 'mayan.apps.mailer.tests.mailers.MailingProfileTest'

TEST_MAILING_PROFILE_LABEL = 'test mailing profile label'
TEST_MAILING_PROFILE_LABEL_EDITED = 'test mailing profile label edited'



from django.db import models

from mayan.apps.testing.tests.base import GenericViewTestCase

from ..classes import ModelCopy
from ..permissions import permission_object_copy

from .literals import TEST_OBJECT_LABEL
from .mixins import ObjectCopyLinkTestMixin


class ObjectCopyLinkTestCase(
    ObjectCopyLinkTestMixin, GenericViewTestCase
):
    auto_create_test_object = True
    auto_create_test_object_fields = {
        'label': models.CharField(max_length=32, unique=True)
    }
    auto_create_test_object_instance_kwargs = {
        'label': TEST_OBJECT_LABEL
    }

    def setUp(self):
        super().setUp()
        ModelCopy(model=self.TestModel, register_permission=True).add_fields(
            field_names=('label',)
        )

    def test_object_copy_link_no_permission(self):
        resolved_link = self._resolve_test_object_copy_link()
        self.assertEqual(resolved_link, None)

    def test_object_copy_link_with_access(self):
        self.grant_access(
            obj=self._test_object, permission=permission_object_copy
        )
        resolved_link = self._resolve_test_object_copy_link()
        self.assertNotEqual(resolved_link, None)



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Redactions'), name='redactions'
)

permission_redaction_create = namespace.add_permission(
    label=_(message='Create new redactions'), name='redaction_create'
)
permission_redaction_delete = namespace.add_permission(
    label=_(message='Delete redactions'), name='redaction_delete'
)
permission_redaction_edit = namespace.add_permission(
    label=_(message='Edit redactions'), name='redaction_edit'
)
permission_redaction_exclude = namespace.add_permission(
    label=_(message='Exclude redactions'), name='redaction_exclude'
)
permission_redaction_view = namespace.add_permission(
    label=_(message='View existing redactions'), name='redaction_view'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.search import (
    search_model_document, search_model_document_version,
    search_model_document_version_page
)

# Document

search_model_document.add_model_field(
    field='versions__version_pages__ocr_content__content',
    label=_(message='Document version OCR')
)

# Document version

search_model_document_version.add_model_field(
    field='version_pages__ocr_content__content', label=_(message='OCR')
)

# Document version page

search_model_document_version_page.add_model_field(
    field='ocr_content__content', label=_(message='Document version OCR')
)



from django.http import Http404
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import SimpleView, SingleObjectListView

from .classes import DependencyGroup
from .forms import DependenciesLicensesForm
from .icons import (
    icon_check_version, icon_dependency_group_list,
    icon_dependency_group_entry_detail, icon_dependency_group_entry_list,
    icon_dependency_licenses
)
from .permissions import permission_dependencies_view
from .utils import PyPIClient


class CheckVersionView(SimpleView):
    template_name = 'appearance/content_container.html'
    view_icon = icon_check_version
    view_permission = permission_dependencies_view

    def get_extra_context(self):
        return {
            'paragraphs': (
                PyPIClient().check_version_verbose(),
                _(
                    'This process only checks the Python component of '
                    'Mayan EDMS.'
                ),
                _(
                    'It does not verify versions of other '
                    'components like packaging or deployment technologies, '
                    'such as container or virtual machine images.'
                ),
            ),
            'title': _(message='Check for updates')
        }


class DependencyGroupEntryListView(SingleObjectListView):
    view_icon = icon_dependency_group_entry_list
    view_permission = permission_dependencies_view

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'object': self.get_object(),
            'subtitle': self.get_object().help_text,
            'title': _(
                'Entries for dependency group: %s'
            ) % self.get_object()
        }

    def get_source_queryset(self):
        return self.get_object().get_entries()

    def get_object(self):
        try:
            return DependencyGroup.get(
                name=self.kwargs['dependency_group_name']
            )
        except KeyError:
            raise Http404(
                _(message='Group %s not found.') % self.kwargs[
                    'dependency_group_name'
                ]
            )


class DependencyGroupListView(SingleObjectListView):
    view_icon = icon_dependency_group_list
    view_permission = permission_dependencies_view

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'title': _(message='Dependency groups')
        }

    def get_source_queryset(self):
        return DependencyGroup.get_all()


class DependencyGroupEntryDetailView(SingleObjectListView):
    view_icon = icon_dependency_group_entry_detail
    view_permission = permission_dependencies_view

    def get_extra_context(self):
        group = self.get_dependency_group()
        entry = self.get_dependency_group_entry()

        return {
            'entry': entry,
            'group': group,
            'hide_link': True,
            'hide_object': True,
            'navigation_object_list': ('group', 'entry'),
            'title': _(
                'Dependency group and entry: %(group)s, %(entry)s'
            ) % {
                'group': group, 'entry': entry
            }
        }

    def get_dependency_group(self):
        try:
            return DependencyGroup.get(
                name=self.kwargs['dependency_group_name']
            )
        except KeyError:
            raise Http404(
                _(message='Group %s not found.') % self.kwargs[
                    'dependency_group_name'
                ]
            )

    def get_dependency_group_entry(self):
        try:
            return self.get_dependency_group().get_entry(
                entry_name=self.kwargs['dependency_group_entry_name']
            )
        except KeyError:
            raise Http404(
                _(message='Entry %s not found.') % self.kwargs[
                    'dependency_group_entry_name'
                ]
            )

    def get_source_queryset(self):
        return self.get_dependency_group_entry().get_dependencies()


class DependencyLicensesView(SimpleView):
    template_name = 'appearance/form_container.html'
    view_icon = icon_dependency_licenses

    def get_extra_context(self):
        # Use a function so that DependenciesLicensesForm get initialized
        # at every request.
        return {
            'form': DependenciesLicensesForm(),
            'read_only': True,
            'title': _(message='Other packages licenses')
        }
